"""Utility objects and functions"""
from __future__ import annotations

from collections.abc import Callable
from collections.abc import Mapping as MappingABC
import copy
from datetime import datetime
from glob import glob
import hashlib
import json
from json.decoder import JSONDecodeError
import logging
import os
import os.path as op
from pathlib import Path
import re
import shutil
import stat
from subprocess import check_output
import sys
import tempfile
from time import sleep
from types import ModuleType
from typing import (
    Any,
    AnyStr,
    Hashable,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)

lgr = logging.getLogger(__name__)

T = TypeVar("T")
V = TypeVar("V")


class SeqInfo(NamedTuple):
    total_files_till_now: int  # 0
    example_dcm_file: str  # 1
    series_id: str  # 2
    dcm_dir_name: str  # 3
    series_files: int  # 4
    unspecified: str  # 5
    dim1: int  # 6
    dim2: int  # 7
    dim3: int  # 8
    dim4: int  # 9
    TR: float  # 10
    TE: float  # 11
    protocol_name: str  # 12
    is_motion_corrected: bool  # 13
    is_derived: bool  # 14
    patient_id: Optional[str]  # 15
    study_description: str  # 16
    referring_physician_name: str  # 17
    series_description: str  # 18
    sequence_name: str  # 19
    image_type: tuple[str, ...]  # 20
    accession_number: str  # 21
    patient_age: Optional[str]  # 22
    patient_sex: Optional[str]  # 23
    date: Optional[str]  # 24
    series_uid: Optional[str]  # 25
    time: Optional[str]  # 26
    custom: Optional[Hashable]  # 27


class StudySessionInfo(NamedTuple):
    # possible prefix identifying the study, e.g.  PI/dataset or just a dataset
    # or empty (default) Note that ATM there should be no multiple DICOMs with
    # the same StudyInstanceUID which would collide, i.e point to the same
    # subject/session. So 'locator' is pretty much an assignment from
    # StudyInstanceUID into some place within hierarchy
    locator: Optional[str]

    session: Optional[str]

    # should be some ID defined either in cmdline or deduced
    subject: Optional[str]


class TempDirs:
    """A helper to centralize handling and cleanup of dirs"""

    def __init__(self) -> None:
        self.dirs: list[str] = []
        self.lgr: logging.Logger = logging.getLogger("tempdirs")

    def __call__(self, prefix: Optional[str] = None) -> str:
        tmpdir = tempfile.mkdtemp(prefix=prefix)
        self.dirs.append(tmpdir)
        return tmpdir

    def __del__(self) -> None:
        try:
            self.cleanup()
        except AttributeError:
            pass

    def exists(self, path: str) -> bool:
        return op.exists(path)

    def cleanup(self) -> None:
        self.lgr.debug("Removing %d temporary directories", len(self.dirs))
        for t in self.dirs[:]:
            self.lgr.debug("Removing %s", t)
            if self:
                self.rmtree(t)
        self.dirs = []

    def rmtree(self, tmpdir: str) -> None:
        if self.exists(tmpdir):
            shutil.rmtree(tmpdir)
        if tmpdir in self.dirs:
            self.dirs.remove(tmpdir)


def docstring_parameter(*sub: str) -> Callable[[T], T]:
    """Borrowed from https://stackoverflow.com/a/10308363/6145776"""

    def dec(obj: T) -> T:
        assert obj.__doc__ is not None
        obj.__doc__ = obj.__doc__.format(*sub)
        return obj

    return dec


def anonymize_sid(sid: AnyStr, anon_sid_cmd: str) -> AnyStr:
    """
    Raises
    ------
    ValueError
      if script returned an empty string (after whitespace stripping),
      or output with multiple words/lines.
    """
    cmd: list[str | bytes] = [anon_sid_cmd, sid]
    shell_return = check_output(cmd)

    if isinstance(shell_return, bytes) and isinstance(sid, str):
        anon_sid = shell_return.decode()
    else:
        anon_sid = shell_return

    anon_sid = anon_sid.strip()
    if not anon_sid:
        raise ValueError(f"{anon_sid_cmd!r} {sid!r} returned empty sid")
    # rudimentary check for sanity: no multiple lines or words (in general
    # ok, but not ok for BIDS) in the output
    if len(anon_sid.split()) > 1:
        raise ValueError(f"{anon_sid_cmd!r} {sid!r} returned multiline output")
    return anon_sid


def create_file_if_missing(filename: str, content: str) -> bool:
    """Create file if missing, so we do not
    override any possibly introduced changes"""
    if op.lexists(filename):
        return False
    dirname = op.dirname(filename)
    if not op.exists(dirname):
        os.makedirs(dirname)
    with open(filename, "w") as f:
        f.write(content)
    return True


def read_config(infile: str) -> dict:
    with open(infile, "r") as fp:
        info = cast(dict, eval(fp.read()))
    return info


def write_config(outfile: str, info: dict) -> None:
    from pprint import PrettyPrinter

    with open(outfile, "w") as fp:
        fp.writelines(PrettyPrinter().pformat(info))


def load_json(filename: str | Path, retry: int = 0) -> Any:
    """Load data from a json file

    Parameters
    ----------
    filename : str or Path
        Filename to load data from.
    retry: int, optional
        Number of times to retry opening/loading the file in case of
        failure.  Code will sleep for 0.1 seconds between retries.
        Could be used in code which is not sensitive to order effects
        (e.g. like populating bids templates where the last one to
        do it, would make sure it would be the correct/final state).

    Returns
    -------
    data : dict
    """
    assert retry >= 0
    for i in range(retry + 1):  # >= 10 sec wait
        try:
            try:
                with open(filename, "r") as fp:
                    data = json.load(fp)
                    break
            except JSONDecodeError:
                lgr.error("{fname} is not a valid json file".format(fname=filename))
                raise
        except (JSONDecodeError, FileNotFoundError) as exc:
            if i >= retry:
                raise
            lgr.warning("Caught %s. Will retry again", exc)
            sleep(0.1)
            continue

    return data


def assure_no_file_exists(path: str | Path) -> None:
    """Check if file or symlink (git-annex?) exists, and if so -- remove"""
    if os.path.lexists(path):
        os.unlink(path)


def save_json(
    filename: str | Path,
    data: Any,
    indent: int = 2,
    sort_keys: bool = True,
    pretty: bool = False,
) -> None:
    """Save data to a json file

    Parameters
    ----------
    filename : str or Path
        Filename to save data in.
    data : dict
        Dictionary to save in json file.
    indent : int, optional
    sort_keys : bool, optional
    pretty : bool, optional

    """
    assure_no_file_exists(filename)
    j: Optional[str] = None
    if pretty:
        try:
            j = json_dumps_pretty(data, sort_keys=sort_keys, indent=indent)
        except AssertionError as exc:
            pretty = False
            lgr.warning(
                "Prettyfication of .json failed (%s).  "
                "Original .json will be kept as is.  Please share (if you "
                "could) "
                "that file (%s) with HeuDiConv developers" % (str(exc), filename)
            )
    if not pretty:
        j = json_dumps(data, sort_keys=sort_keys, indent=indent)
    assert j is not None  # one way or another it should have been set to a str
    with open(filename, "w") as fp:
        fp.write(j)


def json_dumps(json_obj: Any, indent: int = 2, sort_keys: bool = True) -> str:
    """Unified (default indent and sort_keys) invocation of json.dumps"""
    return json.dumps(json_obj, indent=indent, sort_keys=sort_keys)


def json_dumps_pretty(j: Any, indent: int = 2, sort_keys: bool = True) -> str:
    """Given a json structure, pretty print it by colliding numeric arrays
    into a line.

    If resultant structure differs from original -- throws exception
    """
    js = json_dumps(j, indent=indent, sort_keys=sort_keys)
    # trim away \n and spaces between entries of numbers
    js_ = re.sub(
        '[\n ]+("?[-+.0-9e]+"?,?) *\n(?= *"?[-+.0-9e]+"?)',
        r" \1",
        js,
        flags=re.MULTILINE,
    )
    # uniform no spaces before ]
    js_ = re.sub(r" *\]", "]", js_)
    # uniform spacing before numbers
    # But that thing could screw up dates within strings which would have 2 spaces
    # in a date like Mar  3 2017, so we do negative lookahead to avoid changing
    # in those cases
    # import pdb; pdb.set_trace()
    js_ = re.sub(
        r"(?<!\w{3})"  # negative lookbehind for the month
        r'  *("?[-+.0-9e]+"?)'
        r"(?! [123]\d{3})"  # negative lookahead for a year
        r"(?P<space> ?)[ \n]*",
        r" \1\g<space>",
        js_,
    )
    # no spaces after [
    js_ = re.sub(r"\[ ", "[", js_)
    # the load from the original dump and reload from tuned up
    # version should result in identical values since no value
    # must be changed, just formatting.
    j_just_reloaded = json.loads(js)
    j_tuned = json.loads(js_)

    assert j_just_reloaded == j_tuned, (
        "Values differed when they should have not. "
        "Report to the heudiconv developers"
    )

    return js_


def update_json(
    json_file: str | Path, new_data: dict[str, Any], pretty: bool = False
) -> None:
    """
    Adds a given field (and its value) to a json file

    Parameters
    -----------
    json_file : str or Path
        path for the corresponding json file
    new_data : dict
        pair of "key": "value" to add to the json file
    pretty : bool
        argument to be passed to save_json
    """
    for key, value in new_data.items():
        lgr.debug(
            'File "{f}": Setting {k} to {v}'.format(
                f=json_file,
                k=key,
                v=value,
            )
        )

    with open(json_file) as f:
        data = json.load(f)
    data.update(new_data)
    save_json(json_file, data, pretty=pretty)


def treat_infofile(filename: str) -> None:
    """Tune up generated .json file (slim down, pretty-print for humans)."""
    j = load_json(filename)
    j_slim = slim_down_info(j)
    save_json(filename, j_slim, sort_keys=True, pretty=True)
    set_readonly(filename)


def slim_down_info(j: dict[str, Any]) -> dict[str, Any]:
    """Given an aggregated info structure, removes excessive details

    Such as CSA fields, and SourceImageSequence which on Siemens files could be
    huge and not providing any additional immediately usable information.
    If needed, could be recovered from stored DICOMs
    """
    j = copy.deepcopy(j)  # we will do in-place modification on a copy
    dicts: list[dict[str, Any]] = []
    # poor man programming for now
    if "const" in j.get("global", {}):
        dicts.append(j["global"]["const"])
    if "samples" in j.get("time", {}):
        dicts.append(j["time"]["samples"])
    for d in dicts:
        for k in list(d.keys()):
            if k.startswith("Csa") or k.lower() in {"sourceimagesequence"}:
                del d[k]
    return j


def get_known_heuristic_names() -> list[str]:
    """Return a list of heuristic names present under heudiconv/heuristics"""
    import heudiconv.heuristics

    candidates = {
        op.splitext(op.basename(x))[0]
        for hp in heudiconv.heuristics.__path__
        for x in glob(op.join(hp, "*.py")) + glob(op.join(hp, "*.py[co]"))
    }
    return sorted(
        filter(lambda c: not (c.startswith("test_") or c.startswith("_")), candidates)
    )


def load_heuristic(heuristic: str) -> ModuleType:
    """Load heuristic from the file, return the module"""
    if os.path.sep in heuristic or os.path.lexists(heuristic):
        heuristic_file = op.realpath(heuristic)
        path, fname = op.split(heuristic_file)
        try:
            old_syspath = sys.path[:]
            sys.path.insert(0, path)
            mod = __import__(fname.split(".")[0])
            mod.filename = heuristic_file  # type: ignore[attr-defined]
        finally:
            sys.path = old_syspath
    else:
        from importlib import import_module

        try:
            mod = import_module("heudiconv.heuristics.%s" % heuristic)
            assert mod.__file__ is not None
            # remove c or o from pyc/pyo
            mod.filename = mod.__file__.rstrip("co")  # type: ignore[attr-defined]
        except Exception as exc:
            raise ImportError("Failed to import heuristic %s: %s" % (heuristic, exc))
    return mod


def get_heuristic_description(name: str, full: bool = False) -> str:
    try:
        mod = load_heuristic(name)
        desc = (getattr(mod, "__doc__", "") or "").strip()
        return desc.split(os.linesep)[0] if not full else desc
    except Exception as exc:
        return "Failed to load: %s" % exc


def get_known_heuristics_with_descriptions() -> dict[str, str]:
    from collections import OrderedDict

    heuristics = OrderedDict()
    for name in get_known_heuristic_names():
        heuristics[name] = get_heuristic_description(name, full=False)
    return heuristics


def safe_copyfile(src: str, dest: str, overwrite: bool = False) -> None:
    """Copy file but blow if destination name already exists"""
    _safe_op_file(src, dest, "copyfile", overwrite=overwrite)


def safe_movefile(src: str, dest: str, overwrite: bool = False) -> None:
    """Move file but blow if destination name already exists"""
    _safe_op_file(src, dest, "move", overwrite=overwrite)


def _safe_op_file(src: str, dest: str, operation: str, overwrite: bool = False) -> None:
    """Copy or move file but blow if destination name already exists

    Parameters
    ----------
    operation: str, {copyfile, move}
    """
    if op.isdir(dest):
        dest = op.join(dest, op.basename(src))
    if op.realpath(src) == op.realpath(dest):
        lgr.debug("Source %s = destination %s", src, dest)
        return
    if op.lexists(dest):
        if not overwrite:
            raise RuntimeError(
                "was asked to %s %s but destination already exists: %s"
                % (operation, src, dest)
            )
        os.unlink(dest)
    getattr(shutil, operation)(src, dest)


# Globals to check filewriting permissions
ALL_CAN_WRITE = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
ALL_CAN_READ = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
assert ALL_CAN_READ >> 1 == ALL_CAN_WRITE  # Assumption in the code


def set_readonly(path: str, read_only: bool = True) -> int:
    """Make file read only or writeable while preserving "access levels"

    So if file was not readable by others, it should remain not readable by
    others.

    Parameters
    ----------
    path : str
    read_only : bool, optional
        If True (default) - would make it read-only. If False, would make it
        writeable for levels where it is readable

    """

    # get current permissions
    perms = stat.S_IMODE(os.lstat(path).st_mode)
    # set new permissions
    if read_only:
        new_perms = perms & (~ALL_CAN_WRITE)
    else:
        # need to set only for those which had read bit set
        # read bit is <<1 away from write bit
        whocanread = perms & ALL_CAN_READ
        thosecanwrite = whocanread >> 1
        new_perms = perms | thosecanwrite
    # apply and return those target permissions
    os.chmod(path, new_perms)
    return new_perms


def is_readonly(path: str) -> bool:
    """Return True if it is a fully read-only file (dereferences the symlink)"""
    # get current permissions
    perms = stat.S_IMODE(os.lstat(os.path.realpath(path)).st_mode)
    # should be true if anyone is allowed to write
    return not bool(perms & ALL_CAN_WRITE)


def clear_temp_dicoms(item_dicoms: list[str]) -> None:
    """Ensures DICOM temporary directories are safely cleared"""
    try:
        tmp = Path(op.commonprefix(item_dicoms)).parents[1]
    except IndexError:
        return
    if (
        str(tmp.parent) == tempfile.gettempdir()
        and str(tmp.stem).startswith("heudiconvDCM")
        and op.exists(tmp)
    ):
        # clean up directory holding dicoms
        shutil.rmtree(tmp)


def file_md5sum(filename: str) -> str:
    with open(filename, "rb") as f:
        return hashlib.md5(f.read()).hexdigest()


# Borrowed from DataLad (MIT license), with "archives" functionality commented
# out
class File:
    """Helper for a file entry in the create_tree/@with_tree

    It allows to define additional settings for entries
    """

    def __init__(self, name: str, executable: bool = False) -> None:
        """

        Parameters
        ----------
        name : str
          Name of the file
        executable: bool, optional
          Make it executable
        """
        self.name = name
        self.executable = executable

    def __str__(self) -> str:
        return self.name


TreeSpec = Union[
    Sequence[Tuple[Union[str, File], "Load"]],
    Mapping[str, "Load"],
]

Load = Union[str, "TreeSpec", dict]


def create_tree(path: str, tree: TreeSpec, archives_leading_dir: bool = True) -> None:
    """Given a list of tuples (name, load) or a dict create such a tree

    if load is a tuple or a dict itself -- that would create either a subtree
    or an archive with that content and place it into the tree if name ends
    with .tar.gz
    """
    lgr.log(5, "Creating a tree under %s", path)
    if not op.exists(path):
        os.makedirs(path)

    if isinstance(tree, MappingABC):
        tree = list(tree.items())

    for file_, load in tree:
        if isinstance(file_, File):
            executable = file_.executable
            name = file_.name
        else:
            executable = False
            name = file_
        full_name = op.join(path, name)
        if name.endswith(".json") and isinstance(load, dict):
            # (For a json file, we expect the content to be a dictionary, so
            #  don't continue creating a tree, but just write dict to file)
            save_json(full_name, load)
        elif isinstance(load, (tuple, list, dict)):
            # if name.endswith('.tar.gz') or name.endswith('.tar') or name.endswith('.zip'):
            #     create_tree_archive(path, name, load, archives_leading_dir=archives_leading_dir)
            # else:
            create_tree(full_name, load, archives_leading_dir=archives_leading_dir)
        else:
            assert isinstance(load, str)
            with open(full_name, "w") as f:
                f.write(load)
        if executable:
            os.chmod(full_name, os.stat(full_name).st_mode | stat.S_IEXEC)


@overload
def get_typed_attr(obj: Any, attr: str, _type: type[T], default: V) -> T | V:
    ...


@overload
def get_typed_attr(
    obj: Any, attr: str, _type: type[T], default: None = None
) -> T | None:
    ...


def get_typed_attr(
    obj: Any, attr: str, _type: type[T], default: Optional[V] = None
) -> T | V | None:
    """
    Typecasts an object's named attribute. If the attribute cannot be
    converted, the default value is returned instead.

    Parameters
    ----------
    obj: Object
    attr: Attribute
    _type: Type
    default: value, optional
    """
    try:
        val = _type(getattr(obj, attr, default))  # type: ignore[call-arg]
    except (TypeError, ValueError):
        return default
    return val


def get_datetime(date: str, time: str, *, microseconds: bool = True) -> str:
    """
    Combine date and time from dicom to isoformat.

    Parameters
    ----------
    date : str
        Date in YYYYMMDD format.
    time : str
        Time in either HHMMSS.ffffff format or HHMMSS format.
    microseconds: bool, optional
        Either to include microseconds in the output

    Returns
    -------
    datetime_str : str
        Combined date and time in ISO format, with microseconds as
        if fraction was provided in 'time', and 'microseconds' was
        True.
    """
    if "." not in time:
        # add dummy microseconds if not available for strptime to parse
        time += ".000000"
    td = time + ":" + date
    datetime_str = datetime.strptime(td, "%H%M%S.%f:%Y%m%d").isoformat()
    if not microseconds:
        datetime_str = datetime_str.split(".", 1)[0]
    return datetime_str


def strptime_micr(date_string: str, fmt: str) -> datetime:
    r"""
    Decorate strptime while supporting optional [.%f] in the format at the end

    Parameters
    ----------
    date_string: str
      Date string to parse
    fmt: str
      Format string. If it ends with [.%f], we keep it if date_string ends with
      '.\d+' regex and not if it does not.
    """

    optional_micr = "[.%f]"
    if fmt.endswith(optional_micr):
        fmt = fmt[: -len(optional_micr)]
        if re.search(r"\.\d+$", date_string):
            fmt += ".%f"
    return datetime.strptime(date_string, fmt)


def remove_suffix(s: str, suf: str) -> str:
    """
    Remove suffix from the end of the string

    Parameters
    ----------
    s : str
    suf : str

    Returns
    -------
    s : str
        string with "suf" removed from the end (if present)
    """
    if suf and s.endswith(suf):
        return s[: -len(suf)]
    return s


def remove_prefix(s: str, pre: str) -> str:
    """
    Remove prefix from the beginning of the string

    Parameters
    ----------
    s : str
    pre : str

    Returns
    -------
    s : str
        string with "pre" removed from the beginning (if present)
    """
    if pre and s.startswith(pre):
        return s[len(pre) :]
    return s



from __future__ import annotations

import csv
from io import StringIO
import logging
import os
import os.path as op
from os.path import join as opj
from pathlib import Path
import stat
from unittest.mock import patch

import pytest

from heudiconv import __version__
from heudiconv.bids import (
    SCANS_FILE_FIELDS,
    add_participant_record,
    add_rows_to_scans_keys_file,
    find_subj_ses,
    get_formatted_scans_key_row,
    populate_bids_templates,
)
from heudiconv.cli.run import main as runner
from heudiconv.external.dlad import MIN_VERSION, add_to_datalad
from heudiconv.main import process_extra_commands, workflow
from heudiconv.utils import create_file_if_missing, is_readonly, load_json, set_readonly

from .utils import TESTS_DATA_PATH


@patch("sys.stdout", new_callable=StringIO)
def test_main_help(stdout: StringIO) -> None:
    with pytest.raises(SystemExit):
        runner(["--help"])
    assert stdout.getvalue().startswith("usage: ")


@patch("sys.stdout", new_callable=StringIO)
def test_main_version(std: StringIO) -> None:
    with pytest.raises(SystemExit):
        runner(["--version"])
    assert std.getvalue().rstrip() == __version__


def test_create_file_if_missing(tmp_path: Path) -> None:
    tf = tmp_path / "README.txt"
    assert not tf.exists()
    create_file_if_missing(str(tf), "content")
    assert tf.exists()
    assert tf.read_text() == "content"
    create_file_if_missing(str(tf), "content2")
    # nothing gets changed
    assert tf.read_text() == "content"


def test_populate_bids_templates(tmp_path: Path) -> None:
    populate_bids_templates(str(tmp_path), defaults={"Acknowledgements": "something"})
    for f in "README", "dataset_description.json", "CHANGES":
        # Just test that we have created them and they all have stuff TODO
        assert "TODO" in (tmp_path / f).read_text()
    description_file = tmp_path / "dataset_description.json"
    assert "something" in description_file.read_text()

    # it should also be available as a command
    description_file.unlink()

    # it must fail if no heuristic was provided
    with pytest.raises(ValueError) as cme:
        runner(["--command", "populate-templates", "--files", str(tmp_path)])
    assert str(cme.value).startswith("Specify heuristic using -f. Known are:")
    assert "convertall," in str(cme.value)
    assert not description_file.exists()

    runner(
        [
            "--command",
            "populate-templates",
            "-f",
            "convertall",
            "--files",
            str(tmp_path),
        ]
    )
    assert "something" not in description_file.read_text()
    assert "TODO" in description_file.read_text()

    assert load_json(tmp_path / "scans.json") == SCANS_FILE_FIELDS


def test_add_participant_record(tmp_path: Path) -> None:
    tf = tmp_path / "participants.tsv"
    assert not tf.exists()
    add_participant_record(str(tmp_path), "sub01", "023Y", "M")
    # should create the file and place corrected record
    sub01 = tf.read_text()
    assert (
        sub01
        == """\
participant_id	age	sex	group
sub-sub01	23	M	control
"""
    )
    add_participant_record(str(tmp_path), "sub01", "023Y", "F")
    assert tf.read_text() == sub01  # nothing was added even though differs in values
    add_participant_record(str(tmp_path), "sub02", "2", "F")
    assert (
        tf.read_text()
        == """\
participant_id	age	sex	group
sub-sub01	23	M	control
sub-sub02	2	F	control
"""
    )


def test_prepare_for_datalad(tmp_path: Path) -> None:
    pytest.importorskip("datalad", minversion=MIN_VERSION)
    studydir = tmp_path / "PI" / "study"
    studydir_ = str(studydir)
    os.makedirs(studydir_)
    populate_bids_templates(studydir_)

    add_to_datalad(str(tmp_path), studydir_, None, None)

    from datalad.api import Dataset

    superds = Dataset(tmp_path)

    assert superds.is_installed()
    assert not superds.repo.dirty
    subdss = superds.subdatasets(recursive=True, result_xfm="relpaths")
    for ds_path in sorted(subdss):
        ds = Dataset(opj(superds.path, ds_path))
        assert ds.is_installed()
        assert not ds.repo.dirty

    # the last one should have been the study
    target_files = {
        ".bidsignore",
        ".gitattributes",
        ".datalad/config",
        ".datalad/.gitattributes",
        "dataset_description.json",
        "scans.json",
        "CHANGES",
        "README",
    }
    assert set(ds.repo.get_indexed_files()) == target_files
    # and all are under git
    for f in target_files:
        assert not ds.repo.is_under_annex(f)
    assert not ds.repo.is_under_annex(".gitattributes")

    # Above call to add_to_datalad does not create .heudiconv subds since
    # directory does not exist (yet).
    # Let's first check that it is safe to call it again
    add_to_datalad(str(tmp_path), studydir_, None, None)
    assert not ds.repo.dirty

    old_hexsha = ds.repo.get_hexsha()
    # Now let's check that if we had previously converted data so that
    # .heudiconv was not a submodule, we still would not fail
    dsh_path = os.path.join(ds.path, ".heudiconv")
    dummy_path = os.path.join(dsh_path, "dummy.nii.gz")

    create_file_if_missing(dummy_path, "")
    ds.save(dummy_path, message="added a dummy file")
    # next call must not fail, should just issue a warning
    add_to_datalad(str(tmp_path), studydir_, None, None)
    ds.repo.is_under_annex(dummy_path)
    assert not ds.repo.dirty
    assert ".heudiconv/dummy.nii.gz" in ds.repo.get_files()

    # Let's now roll back and make it a proper submodule
    ds.repo.call_git(["reset", "--hard", old_hexsha])
    # now we do not add dummy to git
    create_file_if_missing(dummy_path, "")
    add_to_datalad(str(tmp_path), studydir_, None, None)
    assert ".heudiconv" in ds.subdatasets(result_xfm="relpaths")
    assert not ds.repo.dirty
    assert ".heudiconv/dummy.nii.gz" not in ds.repo.get_files()


def test_get_formatted_scans_key_row() -> None:
    dcm_fn = (
        "%s/01-fmap_acq-3mm/1.3.12.2.1107.5.2.43.66112.2016101409263663466202201.dcm"
        % TESTS_DATA_PATH
    )

    row1 = get_formatted_scans_key_row(dcm_fn)
    assert len(row1) == 3
    assert row1[0] == "2016-10-14T09:26:34.692500"
    assert row1[1] == "n/a"
    prandstr1 = row1[2]

    # if we rerun - should be identical!
    row2 = get_formatted_scans_key_row(dcm_fn)
    prandstr2 = row2[2]
    assert prandstr1 == prandstr2
    assert row1 == row2
    # So it is consistent across pythons etc, we use explicit value here
    assert prandstr1 == "437fe57c"

    # but the prandstr should change when we consider another DICOM file
    row3 = get_formatted_scans_key_row("%s/01-anat-scout/0001.dcm" % TESTS_DATA_PATH)
    assert row3 != row1
    prandstr3 = row3[2]
    assert prandstr1 != prandstr3
    assert prandstr3 == "fae3befb"


# TODO: finish this
def test_add_rows_to_scans_keys_file(tmp_path: Path) -> None:
    fn = opj(tmp_path, "file.tsv")
    rows = {
        "my_file.nii.gz": ["2016adsfasd", "", "fasadfasdf"],
        "another_file.nii.gz": ["2018xxxxx", "", "fasadfasdf"],
    }
    add_rows_to_scans_keys_file(fn, rows)

    def _check_rows(fn: str, rows: dict[str, list[str]]) -> None:
        with open(fn, "r") as csvfile:
            reader = csv.reader(csvfile, delimiter="\t")
            rows_loaded = []
            for row in reader:
                rows_loaded.append(row)
        for i, row_ in enumerate(rows_loaded):
            if i == 0:
                assert row_ == ["filename", "acq_time", "operator", "randstr"]
            else:
                assert rows[row_[0]] == row_[1:]
        # dates, filename should be sorted (date "first", filename "second")
        dates = [(r[1], r[0]) for r in rows_loaded[1:]]
        assert dates == sorted(dates)

    _check_rows(fn, rows)
    # we no longer produce a sidecar .json file there and only generate
    # it while populating templates for BIDS
    assert not op.exists(opj(tmp_path, "file.json"))
    # add a new one
    extra_rows = {
        "a_new_file.nii.gz": ["2016adsfasd23", "", "fasadfasdf"],
        "my_file.nii.gz": ["2016adsfasd", "", "fasadfasdf"],
        "another_file.nii.gz": ["2018xxxxx", "", "fasadfasdf"],
    }
    add_rows_to_scans_keys_file(fn, extra_rows)
    _check_rows(fn, extra_rows)


def test__find_subj_ses() -> None:
    assert find_subj_ses(
        "950_bids_test4/sub-phantom1sid1/fmap/"
        "sub-phantom1sid1_acq-3mm_phasediff.json"
    ) == ("phantom1sid1", None)
    assert find_subj_ses("sub-s1/ses-s1/fmap/sub-s1_ses-s1_acq-3mm_phasediff.json") == (
        "s1",
        "s1",
    )
    assert find_subj_ses("sub-s1/ses-s1/fmap/sub-s1_ses-s1_acq-3mm_phasediff.json") == (
        "s1",
        "s1",
    )
    assert find_subj_ses("fmap/sub-01-fmap_acq-3mm_acq-3mm_phasediff.nii.gz") == (
        "01",
        None,
    )


def test_make_readonly(tmp_path: Path) -> None:
    # we could test it all without torturing a poor file, but for going all
    # the way, let's do it on a file
    path = tmp_path / "f"
    pathname = str(path)
    with open(pathname, "w"):
        pass

    for orig, ro, rw in [
        (0o600, 0o400, 0o600),  # fully returned
        (0o624, 0o404, 0o606),  # it will not get write bit where it is not readable
        (0o1777, 0o1555, 0o1777),  # and other bits should be preserved
    ]:
        os.chmod(pathname, orig)
        assert not is_readonly(pathname)
        assert set_readonly(pathname) == ro
        assert is_readonly(pathname)
        assert stat.S_IMODE(os.lstat(pathname).st_mode) == ro
        # and it should go back if we set it back to non-read_only
        assert set_readonly(pathname, read_only=False) == rw
        assert not is_readonly(pathname)


def test_cache(tmp_path: Path) -> None:
    args = [
        "-f",
        "convertall",
        "--files",
        f"{TESTS_DATA_PATH}/axasc35.dcm",
        "-s",
        "S01",
        "-o",
        str(tmp_path),
    ]
    runner(args)

    cachedir = tmp_path / ".heudiconv" / "S01" / "info"
    assert cachedir.exists()

    # check individual files
    assert (cachedir / "heuristic.py").exists()
    assert (cachedir / "filegroup.json").exists()
    assert (cachedir / "dicominfo.tsv").exists()
    assert (cachedir / "S01.auto.txt").exists()
    assert (cachedir / "S01.edit.txt").exists()

    # check dicominfo has "time" as last column:
    with open(cachedir / "dicominfo.tsv", "r") as f:
        cols = f.readline().split()
    assert cols[26] == "time"


def test_no_etelemetry() -> None:
    # smoke test at large - just verifying that no crash if no etelemetry
    # must not fail if etelemetry no found
    with patch.dict("sys.modules", {"etelemetry": None}):
        workflow(outdir="/dev/null", command="ls", heuristic="reproin", files=[])


# Test two scenarios:
# -study without sessions
# -study with sessions
# The "expected_folder" is the session folder without the tmpdir
@pytest.mark.parametrize(
    "session, expected_folder",
    [("", "foo/sub-{sID}"), ("pre", "foo/sub-{sID}/ses-pre")],
)
def test_populate_intended_for(
    tmp_path: Path, session: str, expected_folder: str, caplog: pytest.LogCaptureFixture
) -> None:
    """
    Tests for "process_extra_commands" when the command is
    'populate-intended-for'
    """
    # Because the function .utils.populate_intended_for already has its own
    # tests, here we just test that "process_extra_commands", when 'command'
    # is 'populate_intended_for' does what we expect (loop through the list of
    # subjects and calls 'populate_intended_for' using the 'POPULATE_INTENDED_FOR_OPTS'
    # defined in the heuristic file 'example.py'). We call it using folders
    # that don't exist, and check that the output is the expected.
    bids_folder = expected_folder.split("sub-")[0]
    subjects = ["1", "2"]
    caplog.set_level(logging.INFO)
    process_extra_commands(
        bids_folder,
        "populate-intended-for",
        [],
        "example",
        session,
        subjects,
        "all",
    )
    for s in subjects:
        expected_info = (
            'Adding "IntendedFor" to the fieldmaps in ' + expected_folder.format(sID=s)
        )
        assert any(expected_info in co.message for co in caplog.records)

    # try the same, but without specifying the subjects or the session.
    # the code in main should find any subject in the output folder and call
    # populate_intended_for on each of them (or for each of the sessions, if
    # the data for that subject is organized in sessions):
    # TODO: Add a 'participants.tsv' file with one of the subjects missing;
    #  the 'process_extra_commands' call should print out a warning
    caplog.clear()
    outdir = opj(tmp_path, bids_folder)
    for subj in subjects:
        subj_dir = opj(outdir, "sub-" + subj)
        print("Creating output dir: %s", subj_dir)
        os.makedirs(subj_dir)
        if session:
            os.makedirs(opj(subj_dir, "ses-" + session))
    process_extra_commands(
        outdir, "populate-intended-for", [], "example", None, [], "all"
    )
    for s in subjects:
        expected_info = 'Adding "IntendedFor" to the fieldmaps in ' + opj(
            tmp_path, expected_folder.format(sID=s)
        )
        assert any(expected_info in co.message for co in caplog.records)



#! /usr/bin/env python3

import hashlib
import re
import sys


def bids_id_(sid: str) -> str:
    m = re.compile(r"^(?:sub-|)(.+)$").search(sid)
    if m:
        parsed_id = m.group(1)
        return hashlib.md5(parsed_id.encode()).hexdigest()[:8]
    else:
        raise ValueError("invalid sid")


def main() -> str:
    sid = sys.argv[1]
    return bids_id_(sid)


if __name__ == "__main__":
    print(main())



from __future__ import annotations

from typing import Optional

from heudiconv.utils import SeqInfo


def create_key(
    template: Optional[str],
    outtype: tuple[str, ...] = ("nii.gz",),
    annotation_classes: None = None,
) -> tuple[str, tuple[str, ...], None]:
    if template is None or not template:
        raise ValueError("Template must be a valid format string")
    return (template, outtype, annotation_classes)


def infotodict(
    seqinfo: list[SeqInfo],
) -> dict[tuple[str, tuple[str, ...], None], list]:
    """Heuristic evaluator for determining which runs belong where

    allowed template fields - follow python string module:

    item: index within category
    subject: participant id
    seqitem: run number during scanning
    subindex: sub index within group
    session: scan index for longitudinal acq
    """
    # for this example, we want to include copies of the DICOMs just for our T1
    # and functional scans
    outdicom = ("dicom", "nii.gz")

    t1 = create_key(
        "{bids_subject_session_dir}/anat/{bids_subject_session_prefix}_T1w",
        outtype=outdicom,
    )
    t2 = create_key("{bids_subject_session_dir}/anat/{bids_subject_session_prefix}_T2w")
    dwi_ap = create_key(
        "{bids_subject_session_dir}/dwi/{bids_subject_session_prefix}_dir-AP_dwi"
    )
    dwi_pa = create_key(
        "{bids_subject_session_dir}/dwi/{bids_subject_session_prefix}_dir-PA_dwi"
    )
    rs = create_key(
        "{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-rest_run-{item:02d}_bold",
        outtype=outdicom,
    )
    boldt1 = create_key(
        "{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-bird1back_run-{item:02d}_bold",
        outtype=outdicom,
    )
    boldt2 = create_key(
        "{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-letter1back_run-{item:02d}_bold",
        outtype=outdicom,
    )
    boldt3 = create_key(
        "{bids_subject_session_dir}/func/{bids_subject_session_prefix}_task-letter2back_run-{item:02d}_bold",
        outtype=outdicom,
    )

    info: dict[tuple[str, tuple[str, ...], None], list] = {
        t1: [],
        t2: [],
        dwi_ap: [],
        dwi_pa: [],
        rs: [],
        boldt1: [],
        boldt2: [],
        boldt3: [],
    }
    for s in seqinfo:
        if (
            (s.dim3 == 176 or s.dim3 == 352)
            and (s.dim4 == 1)
            and ("MEMPRAGE" in s.protocol_name)
        ):
            info[t1] = [s.series_id]
        elif (s.dim4 == 1) and ("MEMPRAGE" in s.protocol_name):
            info[t1] = [s.series_id]
        elif (
            (s.dim3 == 176 or s.dim3 == 352)
            and (s.dim4 == 1)
            and ("T2_SPACE" in s.protocol_name)
        ):
            info[t2] = [s.series_id]
        elif (s.dim4 >= 70) and ("DIFFUSION_HighRes_AP" in s.protocol_name):
            info[dwi_ap].append([s.series_id])
        elif "DIFFUSION_HighRes_PA" in s.protocol_name:
            info[dwi_pa].append([s.series_id])
        elif (s.dim4 == 144) and ("resting" in s.protocol_name):
            if not s.is_motion_corrected:
                info[rs].append([(s.series_id)])
        elif (s.dim4 == 183 or s.dim4 == 366) and ("localizer" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt1].append([s.series_id])
        elif (s.dim4 == 227 or s.dim4 == 454) and ("transfer1" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt2].append([s.series_id])
        elif (s.dim4 == 227 or s.dim4 == 454) and ("transfer2" in s.protocol_name):
            if not s.is_motion_corrected:
                info[boldt3].append([s.series_id])
    return info



#!/usr/bin/env python
from __future__ import annotations

from argparse import ArgumentParser
import logging
import os
import sys
from typing import Optional

from .. import __version__
from ..main import workflow

lgr = logging.getLogger(__name__)


def main(argv: Optional[list[str]] = None) -> None:
    logging.basicConfig(
        format="%(levelname)s: %(message)s",
        level=getattr(logging, os.environ.get("HEUDICONV_LOG_LEVEL", "INFO")),
    )
    parser = get_parser()
    args = parser.parse_args(argv)
    # exit if nothing to be done
    if not args.files and not args.dicom_dir_template and not args.command:
        lgr.warning("Nothing to be done - displaying usage help")
        parser.print_help()
        sys.exit(1)

    kwargs = vars(args)
    workflow(**kwargs)


def get_parser() -> ArgumentParser:
    docstr = """Example:
             heudiconv -d 'rawdata/{subject}' -o . -f heuristic.py -s s1 s2 s3"""
    parser = ArgumentParser(description=docstr)
    parser.add_argument("--version", action="version", version=__version__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-d",
        "--dicom_dir_template",
        dest="dicom_dir_template",
        help="Location of dicomdir that can be indexed with subject id "
        "{subject} and session {session}. Tarballs (can be compressed) "
        "are supported in addition to directory. All matching tarballs "
        "for a subject are extracted and their content processed in a "
        "single pass. If multiple tarballs are found, each is assumed to "
        "be a separate session and the --ses argument is ignored. Note "
        "that you might need to surround the value with quotes to avoid "
        "{...} being considered by shell",
    )
    group.add_argument(
        "--files",
        nargs="*",
        help="Files (tarballs, dicoms) or directories containing files to "
        "process. Cannot be provided if using --dicom_dir_template.",
    )
    parser.add_argument(
        "-s",
        "--subjects",
        dest="subjs",
        type=str,
        nargs="*",
        help="List of subjects - required for dicom template. If not "
        'provided, DICOMS would first be "sorted" and subject IDs '
        "deduced by the heuristic.",
    )
    parser.add_argument(
        "-c",
        "--converter",
        choices=("dcm2niix", "none"),
        default="dcm2niix",
        help='Tool to use for DICOM conversion. Setting to "none" disables '
        "the actual conversion step -- useful for testing heuristics.",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        default=os.getcwd(),
        help="Output directory for conversion setup (for further "
        "customization and future reference. This directory will refer "
        "to non-anonymized subject IDs.",
    )
    parser.add_argument(
        "-l",
        "--locator",
        default=None,
        help="Study path under outdir. If provided, it overloads the value "
        "provided by the heuristic. If --datalad is enabled, every "
        "directory within locator becomes a super-dataset thus "
        'establishing a hierarchy. Setting to "unknown" will skip that '
        "dataset.",
    )
    parser.add_argument(
        "-a",
        "--conv-outdir",
        default=None,
        help="Output directory for converted files. By default this is "
        "identical to --outdir. This option is most useful in "
        "combination with --anon-cmd.",
    )
    parser.add_argument(
        "--anon-cmd",
        default=None,
        help="Command to run to convert subject IDs used for DICOMs to "
        "anonymized IDs. Such command must take a single argument and "
        "return a single anonymized ID. Also see --conv-outdir.",
    )
    parser.add_argument(
        "-f",
        "--heuristic",
        dest="heuristic",
        help="Name of a known heuristic or path to the Python script "
        "containing heuristic.",
    )
    parser.add_argument(
        "-p",
        "--with-prov",
        action="store_true",
        help="Store additional provenance information. Requires python-rdflib.",
    )
    parser.add_argument(
        "-ss",
        "--ses",
        dest="session",
        default=None,
        help="Session for longitudinal study_sessions. Default is None.",
    )
    parser.add_argument(
        "-b",
        "--bids",
        nargs="*",
        metavar=("BIDSOPTION1", "BIDSOPTION2"),
        choices=["notop"],
        dest="bids_options",
        help="Flag for output into BIDS structure. Can also take BIDS-"
        "specific options, e.g., --bids notop. The only currently "
        'supported options is "notop", which skips creation of '
        "top-level BIDS files. This is useful when running in batch "
        "mode to prevent possible race conditions.",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="Overwrite existing converted files.",
    )
    parser.add_argument(
        "--datalad",
        action="store_true",
        help="Store the entire collection as DataLad dataset(s). Small files "
        "will be committed directly to git, while large to annex. New "
        'version (6) of annex repositories will be used in a "thin" '
        "mode so it would look to mortals as just any other regular "
        "directory (i.e. no symlinks to under .git/annex). For now just "
        "for BIDS mode.",
    )
    parser.add_argument(
        "--dbg",
        action="store_true",
        dest="debug",
        help="Do not catch exceptions and show exception traceback.",
    )
    parser.add_argument(
        "--command",
        choices=(
            "heuristics",
            "heuristic-info",
            "ls",
            "populate-templates",
            "sanitize-jsons",
            "treat-jsons",
            "populate-intended-for",
        ),
        help="Custom action to be performed on provided files instead of "
        "regular operation.",
    )
    parser.add_argument(
        "-g",
        "--grouping",
        default="studyUID",
        choices=("studyUID", "accession_number", "all", "custom"),
        help="How to group dicoms (default: by studyUID).",
    )
    parser.add_argument(
        "--minmeta",
        action="store_true",
        help="Exclude dcmstack meta information in sidecar jsons.",
    )
    parser.add_argument(
        "--random-seed", type=int, default=None, help="Random seed to initialize RNG."
    )
    parser.add_argument(
        "--dcmconfig",
        default=None,
        help="JSON file for additional dcm2niix configuration.",
    )
    submission = parser.add_argument_group("Conversion submission options")
    submission.add_argument(
        "-q",
        "--queue",
        choices=("SLURM", None),
        default=None,
        help="Batch system to submit jobs in parallel.",
    )
    submission.add_argument(
        "--queue-args",
        dest="queue_args",
        default=None,
        help="Additional queue arguments passed as a single string of "
        "space-separated Argument=Value pairs.",
    )
    return parser


if __name__ == "__main__":
    main()



from __future__ import annotations

from glob import glob
import os
from os.path import join as opj
from pathlib import Path
import time

from heudiconv.dicoms import compress_dicoms
from heudiconv.utils import TempDirs, file_md5sum

from .utils import TESTS_DATA_PATH


def test_reproducibility(tmp_path: Path) -> None:
    dicom_list = glob(opj(TESTS_DATA_PATH, "01-fmap_acq-3mm", "*"))
    prefix = str(tmp_path / "precious")
    tempdirs = TempDirs()

    tarball = compress_dicoms(dicom_list, prefix, tempdirs, True)
    assert tarball is not None
    md5 = file_md5sum(tarball)
    # must not override, ensure overwrite is set to False
    assert compress_dicoms(dicom_list, prefix, tempdirs, False) is None

    os.unlink(tarball)

    time.sleep(1.1)  # need to guarantee change of time
    tarball_ = compress_dicoms(dicom_list, prefix, tempdirs, True)
    assert tarball == tarball_
    md5_ = file_md5sum(tarball_)
    assert md5 == md5_



from __future__ import annotations

from pathlib import Path
import sys

from nipype.utils.filemanip import which
import pytest

from heudiconv.cli.run import main as runner
from heudiconv.queue import clean_args

from .utils import TESTS_DATA_PATH


@pytest.mark.skipif(bool(which("sbatch")), reason="skip a real slurm call")
@pytest.mark.parametrize(
    "hargs",
    [
        # our new way with automated grouping
        ["--files", f"{TESTS_DATA_PATH}/01-fmap_acq-3mm"],
        # "old" way specifying subject
        ["-d", f"{TESTS_DATA_PATH}/{{subject}}/*", "-s", "01-fmap_acq-3mm"],
    ],
)
def test_queue_no_slurm(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path, hargs: list[str]
) -> None:
    monkeypatch.chdir(tmp_path)
    hargs.extend(["-f", "reproin", "-b", "--minmeta", "--queue", "SLURM"])

    # simulate command-line call
    monkeypatch.setattr(sys, "argv", ["heudiconv"] + hargs)

    with pytest.raises(OSError):  # SLURM should not be installed
        runner(hargs)
    # should have generated a slurm submission script
    slurm_cmd_file = str(tmp_path / "heudiconv-SLURM.sh")
    assert slurm_cmd_file
    # check contents and ensure args match
    with open(slurm_cmd_file) as fp:
        lines = fp.readlines()
    assert lines[0] == "#!/bin/bash\n"
    cmd = lines[1]

    # check that all flags we gave still being called
    for arg in hargs:
        # except --queue <queue>
        if arg in ["--queue", "SLURM"]:
            assert arg not in cmd
        else:
            assert arg in cmd


def test_argument_filtering() -> None:
    cmd_files = [
        "heudiconv",
        "--files",
        "/fake/path/to/files",
        "/another/fake/path",
        "-f",
        "convertall",
        "-q",
        "SLURM",
        "--queue-args",
        "--cpus-per-task=4 --contiguous --time=10",
    ]
    filtered = [
        "heudiconv",
        "--files",
        "/another/fake/path",
        "-f",
        "convertall",
    ]
    assert clean_args(cmd_files, "files", 1) == filtered

    cmd_subjects = [
        "heudiconv",
        "-d",
        "/some/{subject}/path",
        "--queue",
        "SLURM",
        "--subjects",
        "sub1",
        "sub2",
        "sub3",
        "sub4",
        "-f",
        "convertall",
    ]
    filtered = [
        "heudiconv",
        "-d",
        "/some/{subject}/path",
        "--subjects",
        "sub3",
        "-f",
        "convertall",
    ]
    assert clean_args(cmd_subjects, "subjects", 2) == filtered



from __future__ import annotations

from collections.abc import Iterable, Iterator
import os.path as op
from pathlib import Path
import sys
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, Optional, TypeVar
from unittest.mock import patch

import pytest

try:
    from tinydb import Query, TinyDB
except ImportError:
    pytest.importorskip("tinydb")
from subprocess import CalledProcessError

try:
    from heudiconv.cli.monitor import MASK_NEWDIR, monitor, process, run_heudiconv

    class Header(NamedTuple):
        wd: int
        mask: int
        cookie: int
        len: int

    header = Header(5, MASK_NEWDIR, 5, 5)
    watch_path = b"WATCHME"
    filename = b"FILE"
    type_names = b"TYPE"

    path2 = watch_path + b"/" + filename + b"/subpath"

    my_events = [
        (header, type_names, watch_path, filename),
        (header, type_names, path2, b""),
    ]
except AttributeError:
    # Import of inotify fails on mac os x with error
    # lsym(0x11fbeb780, inotify_init): symbol not found
    # because inotify doesn't seem to exist on Mac OS X
    my_events = []
    pytestmark = pytest.mark.skip(reason="Unable to import inotify")


if TYPE_CHECKING:
    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


T = TypeVar("T")


class MockInotifyTree(Generic[T]):
    def __init__(self, events: Iterable[T]) -> None:
        self.events: Iterator[T] = iter(events)

    def event_gen(self) -> Iterator[T]:
        for e in self.events:
            yield e

    def __call__(self, _topdir: Any) -> Self:
        return self


class MockTime:
    def __init__(self, time: float) -> None:
        self.time = time

    def __call__(self) -> float:
        return self.time


@pytest.mark.skip(reason="TODO")
@patch("inotify.adapters.InotifyTree", MockInotifyTree(my_events))
@patch("time.time", MockTime(42))
def test_monitor(capsys: pytest.CaptureFixture[str]) -> None:
    monitor(watch_path.decode(), check_ptrn="")
    out, err = capsys.readouterr()
    desired_output = "{0}/{1} {2}\n".format(watch_path.decode(), filename.decode(), 42)
    desired_output += "Updating {0}/{1}: {2}\n".format(
        watch_path.decode(), filename.decode(), 42
    )
    assert out == desired_output


@pytest.mark.skip(reason="TODO")
@patch("time.time", MockTime(42))
@pytest.mark.parametrize(
    "side_effect,success", [(None, 1), (CalledProcessError(1, "mycmd"), 0)]
)
def test_process(
    tmp_path: Path,
    capsys: pytest.CaptureFixture[str],
    side_effect: Optional[BaseException],
    success: int,
) -> None:
    db_fn = tmp_path / "database.json"
    log_dir = tmp_path / "log"
    log_dir.mkdir()
    db = TinyDB(str(db_fn))
    process_me = "/my/path/A12345"
    accession_number = op.basename(process_me)
    paths2process = {process_me: 42.0}
    with patch("subprocess.Popen") as mocked_popen:
        stdout = b"INFO: PROCESSING STARTS: {'just': 'a test'}"
        mocked_popen_instance = mocked_popen.return_value
        mocked_popen_instance.side_effect = side_effect
        mocked_popen_instance.communicate.return_value = (stdout,)
        # set return value for wait
        mocked_popen_instance.wait.return_value = 1 - success
        # mock also communicate to get the supposed stdout
        process(paths2process, db, wait=-30, logdir=str(log_dir))
        out, err = capsys.readouterr()
        log_fn = log_dir / (accession_number + ".log")

        mocked_popen.assert_called_once()
        assert log_fn.exists()
        assert log_fn.read_text() == stdout.decode("utf-8")
        assert db_fn.exists()
        # dictionary should be empty
        assert not paths2process
        assert out == "Time to process {0}\n".format(process_me)

        # check what we have in the database
        path = Query()
        query = db.get(path.input_path == process_me)
        assert len(db) == 1
        assert query
        assert query["success"] == success
        assert query["accession_number"] == op.basename(process_me)
        assert query["just"] == "a test"


@pytest.mark.skip(reason="TODO")
def test_run_heudiconv() -> None:
    # echo should succeed always
    mydict = {"key1": "value1", "key2": "value2", "success": 1}
    args = ["echo", "INFO:", "PROCESSING", "STARTS:", str(mydict)]
    stdout, info_dict = run_heudiconv(args)
    assert info_dict == mydict
    assert stdout.strip() == " ".join(args[1:])



"""Test functions in heudiconv.bids module.
"""

from __future__ import annotations

from collections import OrderedDict
from collections.abc import Callable
from datetime import datetime, timedelta
from glob import glob
import itertools
import os
import os.path as op
from pathlib import Path
from random import choice, random, seed, shuffle
import re
import string
from typing import Any, Dict, List, Optional, Tuple

import nibabel
from numpy import testing as np_testing
import pytest

from heudiconv.bids import (
    SHIM_KEY,
    AllowedCriteriaForFmapAssignment,
    BIDSFile,
    KeyInfoForForce,
    find_compatible_fmaps_for_run,
    find_compatible_fmaps_for_session,
    find_fmap_groups,
    get_key_info_for_fmap_assignment,
    get_shim_setting,
    maybe_na,
    populate_intended_for,
    sanitize_label,
    select_fmap_from_compatible_groups,
    treat_age,
)
from heudiconv.cli.run import main as runner
from heudiconv.utils import Load, create_tree, load_json, remove_suffix, save_json

from .utils import TESTS_DATA_PATH, fetch_data, gen_heudiconv_args

have_datalad = True
try:
    from datalad.support.exceptions import IncompleteResultsError
except ImportError:
    have_datalad = False


def gen_rand_label(label_size: int, label_seed: int, seed_stdout: bool = True) -> str:
    seed(label_seed)
    rand_char = "".join(choice(string.ascii_letters) for _ in range(label_size - 1))
    seed(label_seed)
    rand_num = choice(string.digits)
    if seed_stdout:
        print(f"Seed used to generate custom label: {label_seed}")
    return rand_char + rand_num


def test_maybe_na() -> None:
    for na in "", " ", None, "n/a", "N/A", "NA":
        assert maybe_na(na) == "n/a"
    for notna in 0, 1, False, True, "value":
        assert maybe_na(notna) == str(notna)


def test_treat_age() -> None:
    assert treat_age(None) is None
    assert treat_age(0) == "0"
    assert treat_age("0") == "0"
    assert treat_age("0000") == "0"
    assert treat_age("0000Y") == "0"
    assert treat_age("000.1Y") == "0.1"
    assert treat_age("1M") == "0.08"
    assert treat_age("12M") == "1"
    assert treat_age("0000.1") == "0.1"
    assert treat_age(0000.1) == "0.1"


SHIM_LENGTH = 6
TODAY = datetime.today()
LABEL_SEED = int.from_bytes(os.urandom(8), byteorder="big")

A_SHIM = [random() for i in range(SHIM_LENGTH)]


def test_get_shim_setting(tmp_path: Path) -> None:
    """Tests for get_shim_setting"""
    json_dir = op.join(tmp_path, "foo")
    if not op.exists(json_dir):
        os.makedirs(json_dir)
    json_name = op.join(json_dir, "sub-foo.json")
    # 1) file with no "ShimSetting", should return None
    save_json(json_name, {})
    with pytest.raises(KeyError):
        assert get_shim_setting(json_name)

    # -file with "ShimSetting" field
    save_json(json_name, {SHIM_KEY: A_SHIM})
    assert get_shim_setting(json_name) == A_SHIM


def test_get_key_info_for_fmap_assignment(
    tmp_path: Path, label_size: int = 4, label_seed: int = LABEL_SEED
) -> None:
    """
    Test get_key_info_for_fmap_assignment.

    label_size and label_seed are used for the "CustomAcquisitionLabel" matching
    parameter. label_size is the size of the random label while label_seed is
    the seed for the random label creation.
    """

    nifti_file = op.join(TESTS_DATA_PATH, "sample_nifti.nii.gz")
    # Get the expected parameters from the NIfTI header:
    MY_HEADER = nibabel.ni1.np.loadtxt(
        op.join(TESTS_DATA_PATH, remove_suffix(nifti_file, ".nii.gz") + "_params.txt")
    )
    json_name = op.join(TESTS_DATA_PATH, remove_suffix(nifti_file, ".nii.gz") + ".json")

    # 1) Call for a non-existing file should give an error:
    with pytest.raises(FileNotFoundError):
        assert get_key_info_for_fmap_assignment("foo.json", "ImagingVolume")

    # 2) matching_parameters = 'Shims'
    json_name = op.join(TESTS_DATA_PATH, remove_suffix(nifti_file, ".nii.gz") + ".json")
    save_json(
        json_name, {SHIM_KEY: A_SHIM}
    )  # otherwise get_key_info_for_fmap_assignment will give an error
    key_info = get_key_info_for_fmap_assignment(json_name, matching_parameter="Shims")
    assert key_info == [A_SHIM]

    # 3) matching_parameters = 'ImagingVolume'
    key_info = get_key_info_for_fmap_assignment(
        json_name, matching_parameter="ImagingVolume"
    )
    np_testing.assert_almost_equal(key_info[0], MY_HEADER[:4], decimal=6)
    np_testing.assert_almost_equal(key_info[1], MY_HEADER[4][:3], decimal=6)

    # 4) matching_parameters = 'Force'
    key_info = get_key_info_for_fmap_assignment(json_name, matching_parameter="Force")
    assert key_info == [KeyInfoForForce]

    # 5) matching_parameter = 'ModalityAcquisitionLabel'
    for d in ["fmap", "func", "dwi", "anat"]:
        (tmp_path / d).mkdir(parents=True, exist_ok=True)
    for dirname, fname, expected_key_info in [
        ("fmap", "sub-foo_acq-fmri_epi.json", "func"),
        ("fmap", "sub-foo_acq-bold_epi.json", "func"),
        ("fmap", "sub-foo_acq-func_epi.json", "func"),
        ("fmap", "sub-foo_acq-diff_epi.json", "dwi"),
        ("fmap", "sub-foo_acq-anat_epi.json", "anat"),
        ("fmap", "sub-foo_acq-struct_epi.json", "anat"),
        ("func", "sub-foo_bold.json", "func"),
        ("dwi", "sub-foo_dwi.json", "dwi"),
        ("anat", "sub-foo_T1w.json", "anat"),
    ]:
        json_name = op.join(tmp_path, dirname, fname)
        save_json(json_name, {SHIM_KEY: A_SHIM})
        assert [expected_key_info] == get_key_info_for_fmap_assignment(
            json_name, matching_parameter="ModalityAcquisitionLabel"
        )

    # 6) matching_parameter = 'CustomAcquisitionLabel'
    A_LABEL = gen_rand_label(label_size, label_seed)
    for d in ["fmap", "func", "dwi", "anat"]:
        (tmp_path / d).mkdir(parents=True, exist_ok=True)

    for dirname, fname, expected_key_info in [
        ("fmap", f"sub-foo_acq-{A_LABEL}_epi.json", A_LABEL),
        ("func", f"sub-foo_task-{A_LABEL}_acq-foo_bold.json", A_LABEL),
        ("dwi", f"sub-foo_acq-{A_LABEL}_dwi.json", A_LABEL),
        ("anat", f"sub-foo_acq-{A_LABEL}_T1w.json", A_LABEL),
    ]:
        json_name = op.join(tmp_path, dirname, fname)
        save_json(json_name, {SHIM_KEY: A_SHIM})
        assert [expected_key_info] == get_key_info_for_fmap_assignment(
            json_name, matching_parameter="CustomAcquisitionLabel"
        )

    # Finally: invalid matching_parameters:
    assert (
        get_key_info_for_fmap_assignment(json_name, matching_parameter="Invalid") == []
    )


def generate_scans_tsv(session_struct: dict[str, dict[str, Any]]) -> str:
    """
    Generates the contents of the "_scans.tsv" file, given a session structure.
    Currently, it will have the columns "filename" and "acq_time".
    The acq_time will increase by one minute from run to run.

    Parameters
    ----------
    session_struct : dict
        structure for the session, as a dict with modality: files

    Returns
    -------
    scans_file_content : str
        multi-line string with the content of the file
    """
    # for each modality in session_struct (k), get the filenames:
    scans_fnames = [
        op.join(k, vk)
        for k, v in session_struct.items()
        for vk in sorted(v.keys())
        if vk.endswith(".nii.gz")
    ]
    # for each file, increment the acq_time by one minute:
    scans_file_content = ["filename\tacq_time"] + [
        "%s\t%s" % (fn, (TODAY + timedelta(minutes=i)).isoformat())
        for fn, i in zip(scans_fnames, range(len(scans_fnames)))
    ]
    # convert to multiline string:
    return "\n".join(scans_file_content)


DummySession = Tuple[
    Dict[str, Load],
    Dict[str, Optional[List[str]]],
    Dict[str, List[str]],
    Dict[str, Dict[str, List[str]]],
]


def create_dummy_pepolar_bids_session(session_path: str) -> DummySession:
    """
    Creates a dummy BIDS session, with slim json files and empty nii.gz
    The fmap files are pepolar
    The json files have ShimSettings

    Parameters
    ----------
    session_path : str
        path to the session (or subject) level folder

    Returns
    -------
    session_struct : dict
        Structure of the directory that was created
    expected_result : dict
        dictionary with fmap names as keys and the expected "IntendedFor" as
        values.
    expected_fmap_groups : dict
        dictionary with the expected fmap groups
    expected_compatible_fmaps : dict
        dictionary with the expected fmap groups for each non-fmap run in the
        session
    """
    session_parent, session_basename = op.split(session_path.rstrip(op.sep))
    if session_basename.startswith("ses-"):
        prefix = op.split(session_parent)[1] + "_" + session_basename
    else:
        prefix = session_basename

    # 1) Simulate the file structure for a session:

    # Generate some random ShimSettings:
    anat_shims = [random() for i in range(SHIM_LENGTH)]
    dwi_shims = [random() for i in range(SHIM_LENGTH)]
    func_shims_A = [random() for i in range(SHIM_LENGTH)]
    func_shims_B = [random() for i in range(SHIM_LENGTH)]

    # Dict with the file structure for the session:
    # -anat:
    anat_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_{m}.nii.gz".format(p=prefix, m=mod): "" for mod in ["T1w", "T2w"]
    }
    anat_struct.update(
        {
            "{p}_{m}.json".format(p=prefix, m=mod): {"ShimSetting": anat_shims}
            for mod in ["T1w", "T2w"]
        }
    )
    # -dwi:
    dwi_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_acq-A_run-{r}_dwi.nii.gz".format(p=prefix, r=runNo): "" for runNo in [1, 2]
    }
    dwi_struct.update(
        {
            "{p}_acq-A_run-{r}_dwi.json".format(p=prefix, r=runNo): {
                "ShimSetting": dwi_shims
            }
            for runNo in [1, 2]
        }
    )
    # -func:
    func_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_acq-{a}_bold.nii.gz".format(p=prefix, a=acq): ""
        for acq in ["A", "B", "unmatched"]
    }
    func_struct.update(
        {
            "{p}_acq-A_bold.json".format(p=prefix): {"ShimSetting": func_shims_A},
            "{p}_acq-B_bold.json".format(p=prefix): {"ShimSetting": func_shims_B},
            "{p}_acq-unmatched_bold.json".format(p=prefix): {
                "ShimSetting": [random() for i in range(SHIM_LENGTH)]
            },
        }
    )
    # -fmap:
    #  * NIfTI files:
    fmap_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_acq-{a}_dir-{d}_run-{r}_epi.nii.gz".format(p=prefix, a=acq, d=d, r=r): ""
        for acq in ["dwi", "fMRI"]
        for d in ["AP", "PA"]
        for r in [1, 2]
    }
    #  * dwi shims:
    expected_fmap_groups = {
        "{p}_acq-dwi_run-{r}_epi".format(p=prefix, r=r): [
            "{p}_acq-dwi_dir-{d}_run-{r}_epi.json".format(
                p=op.join(session_path, "fmap", prefix), d=d, r=r
            )
            for d in ["AP", "PA"]
        ]
        for r in [1, 2]
    }
    fmap_struct.update(
        {
            "{p}_acq-dwi_dir-{d}_run-{r}_epi.json".format(p=prefix, d=d, r=r): {
                "ShimSetting": dwi_shims
            }
            for d in ["AP", "PA"]
            for r in [1, 2]
        }
    )
    #  * func_shims (_A and _B):
    expected_fmap_groups.update(
        {
            "{p}_acq-fMRI_run-{r}_epi".format(p=prefix, r=r): [
                "{p}_acq-fMRI_dir-{d}_run-{r}_epi.json".format(
                    p=op.join(session_path, "fmap", prefix), d=d, r=r
                )
                for d in ["AP", "PA"]
            ]
            for r in [1, 2]
        }
    )
    fmap_struct.update(
        {
            "{p}_acq-fMRI_dir-{d}_run-{r}_epi.json".format(p=prefix, d=d, r=r): {
                "ShimSetting": shims
            }
            for r, shims in {"1": func_shims_A, "2": func_shims_B}.items()
            for d in ["AP", "PA"]
        }
    )
    # structure for the full session (init the OrderedDict as a list to preserve order):
    session_struct = {
        "fmap": fmap_struct,
        "anat": anat_struct,
        "dwi": dwi_struct,
        "func": func_struct,
    }
    # add "_scans.tsv" file to the session_struct
    scans_file_content = generate_scans_tsv(session_struct)
    session_struct2: dict[str, Load] = {
        **session_struct,
        f"{prefix}_scans.tsv": scans_file_content,
    }

    create_tree(session_path, session_struct2)

    # 2) Now, let's create a dict with the fmap groups compatible for each run
    # -anat: empty
    expected_compatible_fmaps: dict[str, dict[str, list[str]]] = {
        "{p}_{m}.json".format(p=op.join(session_path, "anat", prefix), m=mod): {}
        for mod in ["T1w", "T2w"]
    }
    # -dwi: each of the runs (1, 2) is compatible with both of the dwi fmaps (1, 2):
    expected_compatible_fmaps.update(
        {
            "{p}_acq-A_run-{r}_dwi.json".format(
                p=op.join(session_path, "dwi", prefix), r=runNo
            ): {
                key: val
                for key, val in expected_fmap_groups.items()
                if key
                in ["{p}_acq-dwi_run-{r}_epi".format(p=prefix, r=r) for r in [1, 2]]
            }
            for runNo in [1, 2]
        }
    )
    # -func: acq-A is compatible w/ fmap fMRI run 1; acq-2 w/ fmap fMRI run 2
    expected_compatible_fmaps.update(
        {
            "{p}_acq-{a}_bold.json".format(
                p=op.join(session_path, "func", prefix), a=acq
            ): {
                key: val
                for key, val in expected_fmap_groups.items()
                if key in ["{p}_acq-fMRI_run-{r}_epi".format(p=prefix, r=runNo)]
            }
            for runNo, acq in {"1": "A", "2": "B"}.items()
        }
    )
    # -func (cont): acq-unmatched is empty
    expected_compatible_fmaps.update(
        {
            "{p}_acq-unmatched_bold.json".format(
                p=op.join(session_path, "func", prefix)
            ): {}
        }
    )

    # 3) Then, let's create a dict with what we expect for the "IntendedFor":

    sub_match = re.findall("(sub-([a-zA-Z0-9]*))", session_path)
    sub_str = sub_match[0][0]
    expected_prefix = session_path.split(sub_str)[-1].split(op.sep)[-1]

    # dict, with fmap names as keys and the expected "IntendedFor" as values.
    expected_result = {
        "{p}_acq-dwi_dir-{d}_run-{r}_epi.json".format(
            p=prefix, d=d, r=runNo
        ): intended_for
        # (runNo=1 goes with the long list, runNo=2 goes with None):
        for runNo, intended_for in zip(
            [1, 2],
            [
                [
                    op.join(
                        expected_prefix,
                        "dwi",
                        "{p}_acq-A_run-{r}_dwi.nii.gz".format(p=prefix, r=r),
                    )
                    for r in [1, 2]
                ],
                None,
            ],
        )
        for d in ["AP", "PA"]
    }
    expected_result.update(
        {
            "{p}_acq-fMRI_dir-{d}_run-{r}_epi.json".format(p=prefix, d=d, r=runNo): [
                op.join(
                    expected_prefix,
                    "func",
                    "{p}_acq-{a}_bold.nii.gz".format(p=prefix, a=acq),
                )
            ]
            # runNo=1 goes with acq='A'; runNo=2 goes with acq='B'
            for runNo, acq in zip([1, 2], ["A", "B"])
            for d in ["AP", "PA"]
        }
    )

    return (
        session_struct2,
        expected_result,
        expected_fmap_groups,
        expected_compatible_fmaps,
    )


def create_dummy_no_shim_settings_bids_session(session_path: str) -> DummySession:
    """
    Creates a dummy BIDS session, with slim json files and empty nii.gz
    The fmap files are pepolar
    The json files don't have ShimSettings

    Parameters
    ----------
    session_path : str
        path to the session (or subject) level folder

    Returns
    -------
    session_struct : dict
        Structure of the directory that was created
    expected_result : dict
        dictionary with fmap names as keys and the expected "IntendedFor" as
        values.
    None
        it returns a third argument (None) to have the same signature as
        create_dummy_pepolar_bids_session
    """
    session_parent, session_basename = op.split(session_path.rstrip(op.sep))
    if session_basename.startswith("ses-"):
        prefix = op.split(session_parent)[1] + "_" + session_basename
    else:
        prefix = session_basename

    # 1) Simulate the file structure for a session:

    # Dict with the file structure for the session.
    # All json files will be empty.
    # -anat:
    anat_struct = mkstruct(f"{prefix}_{{0}}.{{ext}}", ["T1w", "T2w"])
    # -dwi:
    dwi_struct = mkstruct(f"{prefix}_acq-A_run-{{0}}_dwi.{{ext}}", [1, 2])
    # -func:
    func_struct = mkstruct(f"{prefix}_acq-{{0}}_bold.{{ext}}", ["A", "B"])
    # -fmap:
    fmap_struct = mkstruct(
        f"{prefix}_acq-{{0}}_dir-{{1}}_run-{{2}}_epi.{{ext}}",
        ["dwi", "fMRI"],
        ["AP", "PA"],
        [1, 2],
    )
    expected_fmap_groups = {
        "{p}_acq-{a}_run-{r}_epi".format(p=prefix, a=acq, r=r): [
            "{p}_acq-{a}_dir-{d}_run-{r}_epi.json".format(
                p=op.join(session_path, "fmap", prefix), a=acq, d=d, r=r
            )
            for d in ["AP", "PA"]
        ]
        for acq in ["dwi", "fMRI"]
        for r in [1, 2]
    }

    # structure for the full session (init the OrderedDict as a list to preserve order):
    session_struct = OrderedDict(
        [
            ("fmap", fmap_struct),
            ("anat", anat_struct),
            ("dwi", dwi_struct),
            ("func", func_struct),
        ]
    )
    # add "_scans.tsv" file to the session_struct
    scans_file_content = generate_scans_tsv(session_struct)
    session_struct2: dict[str, Load] = {
        **session_struct,
        f"{prefix}_scans.tsv": scans_file_content,
    }

    create_tree(session_path, session_struct2)

    # 2) Now, let's create a dict with the fmap groups compatible for each run
    # -anat: empty
    expected_compatible_fmaps: dict[str, dict[str, list[str]]] = {
        "{p}_{m}.json".format(p=op.join(session_path, "anat", prefix), m=mod): {}
        for mod in ["T1w", "T2w"]
    }
    # -dwi: each of the runs (1, 2) is compatible with both of the dwi fmaps (1, 2):
    expected_compatible_fmaps.update(
        {
            "{p}_acq-A_run-{r}_dwi.json".format(
                p=op.join(session_path, "dwi", prefix), r=runNo
            ): {
                key: val
                for key, val in expected_fmap_groups.items()
                if key
                in ["{p}_acq-dwi_run-{r}_epi".format(p=prefix, r=r) for r in [1, 2]]
            }
            for runNo in [1, 2]
        }
    )
    # -func: each of the acq (A, B) is compatible w/ both fmap fMRI runs (1, 2)
    expected_compatible_fmaps.update(
        {
            "{p}_acq-{a}_bold.json".format(
                p=op.join(session_path, "func", prefix), a=acq
            ): {
                key: val
                for key, val in expected_fmap_groups.items()
                if key
                in ["{p}_acq-fMRI_run-{r}_epi".format(p=prefix, r=r) for r in [1, 2]]
            }
            for acq in ["A", "B"]
        }
    )

    # 3) Now, let's create a dict with what we expect for the "IntendedFor":
    # NOTE: The "expected_prefix" (the beginning of the path to the
    # "IntendedFor") should be relative to the subject level (see:
    # https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/01-magnetic-resonance-imaging-data.html#fieldmap-data)

    sub_match = re.findall("(sub-([a-zA-Z0-9]*))", session_path)
    sub_str = sub_match[0][0]
    expected_prefix = session_path.split(sub_str)[-1].split(op.sep)[-1]

    # dict, with fmap names as keys and the expected "IntendedFor" as values.
    expected_result = {
        # (runNo=1 goes with the long list, runNo=2 goes with None):
        "{p}_acq-dwi_dir-{d}_run-{r}_epi.json".format(
            p=prefix, d=d, r=runNo
        ): intended_for
        for runNo, intended_for in zip(
            [1, 2],
            [
                [
                    op.join(
                        expected_prefix,
                        "dwi",
                        "{p}_acq-A_run-{r}_dwi.nii.gz".format(p=prefix, r=r),
                    )
                    for r in [1, 2]
                ],
                None,
            ],
        )
        for d in ["AP", "PA"]
    }
    expected_result.update(
        {
            # The first "fMRI" run gets all files in the "func" folder;
            # the second shouldn't get any.
            "{p}_acq-fMRI_dir-{d}_run-{r}_epi.json".format(
                p=prefix, d=d, r=runNo
            ): intended_for
            for runNo, intended_for in zip(
                [1, 2],
                [
                    [
                        op.join(
                            expected_prefix,
                            "func",
                            "{p}_acq-{a}_bold.nii.gz".format(p=prefix, a=acq),
                        )
                        for acq in ["A", "B"]
                    ],
                    None,
                ],
            )
            for d in ["AP", "PA"]
        }
    )

    return (
        session_struct2,
        expected_result,
        expected_fmap_groups,
        expected_compatible_fmaps,
    )


def create_dummy_no_shim_settings_custom_label_bids_session(
    session_path: str, label_size: int = 4, label_seed: int = LABEL_SEED
) -> DummySession:
    """
    Creates a dummy BIDS session, with slim json files and empty nii.gz
    The fmap files are pepolar
    The json files don't have ShimSettings
    The fmap files have a custom ACQ label matching:
        - TASK label for <func> modality
        - ACQ label for any other modality (e.g. <dwi>)

    Parameters
    ----------
    session_path : str
        path to the session (or subject) level folder
    label_size : int, optional
        size of the random label
    label_seed : int, optional
        seed for the random label creation

    Returns
    -------
    session_struct : dict
        Structure of the directory that was created
    expected_result : dict
        dictionary with fmap names as keys and the expected "IntendedFor" as
        values.
    None
        it returns a third argument (None) to have the same signature as
        create_dummy_pepolar_bids_session
    """
    session_parent, session_basename = op.split(session_path.rstrip(op.sep))
    if session_basename.startswith("ses-"):
        prefix = op.split(session_parent)[1] + "_" + session_basename
    else:
        prefix = session_basename

    # 1) Simulate the file structure for a session:

    # Dict with the file structure for the session.
    # All json files will be empty.
    # -anat:
    anat_struct = mkstruct(f"{prefix}_{{0}}.{{ext}}", ["T1w", "T2w"])
    # -dwi:
    label_seed += 1
    DWI_LABEL = gen_rand_label(label_size, label_seed)
    dwi_struct = mkstruct(f"{prefix}_acq-{DWI_LABEL}_run-{{0}}_dwi.{{ext}}", [1, 2])
    # -func:
    label_seed += 1
    FUNC_LABEL = gen_rand_label(label_size, label_seed)
    func_struct = mkstruct(
        f"{prefix}_task-{FUNC_LABEL}_acq-{{0}}_bold.{{ext}}",
        ["A", "B"],
    )
    # -fmap:
    fmap_struct = mkstruct(
        f"{prefix}_acq-{{0}}_dir-{{1}}_run-{{2}}_epi.{{ext}}",
        [DWI_LABEL, FUNC_LABEL],
        ["AP", "PA"],
        [1, 2],
    )
    expected_fmap_groups = {
        f"{prefix}_acq-{acq}_run-{r}_epi": [
            f'{op.join(session_path, "fmap", prefix)}_acq-{acq}_dir-{d}_run-{r}_epi.json'
            for d in ["AP", "PA"]
        ]
        for acq in [DWI_LABEL, FUNC_LABEL]
        for r in [1, 2]
    }

    # structure for the full session (init the OrderedDict as a list to preserve order):
    session_struct = OrderedDict(
        [
            ("fmap", fmap_struct),
            ("anat", anat_struct),
            ("dwi", dwi_struct),
            ("func", func_struct),
        ]
    )
    # add "_scans.tsv" file to the session_struct
    scans_file_content = generate_scans_tsv(session_struct)
    session_struct2: dict[str, Load] = {
        **session_struct,
        f"{prefix}_scans.tsv": scans_file_content,
    }

    create_tree(session_path, session_struct2)

    # 2) Now, let's create a dict with the fmap groups compatible for each run
    # -anat: empty
    expected_compatible_fmaps: dict[str, dict[str, list[str]]] = {
        f'{op.join(session_path, "anat", prefix)}_{mod}.json': {}
        for mod in ["T1w", "T2w"]
    }
    # -dwi: each of the runs (1, 2) is compatible with both of the dwi fmaps (1, 2):
    expected_compatible_fmaps.update(
        {
            f'{op.join(session_path, "dwi", prefix)}_acq-{DWI_LABEL}_run-{runNo}_dwi.json': {
                key: val
                for key, val in expected_fmap_groups.items()
                if key in [f"{prefix}_acq-{DWI_LABEL}_run-{r}_epi" for r in [1, 2]]
            }
            for runNo in [1, 2]
        }
    )
    # -func: each of the acq (A, B) is compatible w/ both fmap fMRI runs (1, 2)
    expected_compatible_fmaps.update(
        {
            f'{op.join(session_path, "func", prefix)}_task-{FUNC_LABEL}_acq-{acq}_bold.json': {
                key: val
                for key, val in expected_fmap_groups.items()
                if key in [f"{prefix}_acq-{FUNC_LABEL}_run-{r}_epi" for r in [1, 2]]
            }
            for acq in ["A", "B"]
        }
    )

    # 3) Now, let's create a dict with what we expect for the "IntendedFor":
    # NOTE: The "expected_prefix" (the beginning of the path to the
    # "IntendedFor") should be relative to the subject level (see:
    # https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/01-magnetic-resonance-imaging-data.html#fieldmap-data)

    sub_match = re.findall("(sub-([a-zA-Z0-9]*))", session_path)
    sub_str = sub_match[0][0]
    expected_prefix = session_path.split(sub_str)[-1].split(op.sep)[-1]

    # dict, with fmap names as keys and the expected "IntendedFor" as values.
    expected_result = {
        # (runNo=1 goes with the long list, runNo=2 goes with None):
        f"{prefix}_acq-{DWI_LABEL}_dir-{d}_run-{runNo}_epi.json": intended_for
        for runNo, intended_for in zip(
            [1, 2],
            [
                [
                    op.join(
                        expected_prefix,
                        "dwi",
                        f"{prefix}_acq-{DWI_LABEL}_run-{r}_dwi.nii.gz",
                    )
                    for r in [1, 2]
                ],
                None,
            ],
        )
        for d in ["AP", "PA"]
    }
    expected_result.update(
        {
            # The first "fMRI" run gets all files in the "func" folder;
            # the second shouldn't get any.
            f"{prefix}_acq-{FUNC_LABEL}_dir-{d}_run-{runNo}_epi.json": intended_for
            for runNo, intended_for in zip(
                [1, 2],
                [
                    [
                        op.join(
                            expected_prefix,
                            "func",
                            f"{prefix}_task-{FUNC_LABEL}_acq-{acq}_bold.nii.gz",
                        )
                        for acq in ["A", "B"]
                    ],
                    None,
                ],
            )
            for d in ["AP", "PA"]
        }
    )

    return (
        session_struct2,
        expected_result,
        expected_fmap_groups,
        expected_compatible_fmaps,
    )


def create_dummy_magnitude_phase_bids_session(session_path: str) -> DummySession:
    """
    Creates a dummy BIDS session, with slim json files and empty nii.gz
    The fmap files are a magnitude/phase pair
    The json files have ShimSettings
    We just need to test a very simple case to make sure the mag/phase have
    the same "IntendedFor" field:

    Parameters
    ----------
    session_path : str
        path to the session (or subject) level folder

    Returns
    -------
    session_struct : dict
        Structure of the directory that was created
    expected_result : dict
        dictionary with fmap names as keys and the expected "IntendedFor" as
        values.
    expected_fmap_groups : dict
        dictionary with the expected fmap groups
    """
    session_parent, session_basename = op.split(session_path.rstrip(op.sep))
    if session_basename.startswith("ses-"):
        prefix = op.split(session_parent)[1] + "_" + session_basename
    else:
        prefix = session_basename

    # 1) Simulate the file structure for a session:

    # Generate some random ShimSettings:
    dwi_shims = [random() for i in range(SHIM_LENGTH)]
    func_shims_A = [random() for i in range(SHIM_LENGTH)]
    func_shims_B = [random() for i in range(SHIM_LENGTH)]

    # Dict with the file structure for the session:
    # -dwi:
    dwi_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_acq-A_run-{r}_dwi.nii.gz".format(p=prefix, r=runNo): "" for runNo in [1, 2]
    }
    dwi_struct.update(
        {
            "{p}_acq-A_run-{r}_dwi.json".format(p=prefix, r=runNo): {
                "ShimSetting": dwi_shims
            }
            for runNo in [1, 2]
        }
    )
    # -func:
    func_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_acq-{a}_bold.nii.gz".format(p=prefix, a=acq): ""
        for acq in ["A", "B", "unmatched"]
    }
    func_struct.update(
        {
            "{p}_acq-A_bold.json".format(p=prefix): {"ShimSetting": func_shims_A},
            "{p}_acq-B_bold.json".format(p=prefix): {"ShimSetting": func_shims_B},
            "{p}_acq-unmatched_bold.json".format(p=prefix): {
                "ShimSetting": [random() for i in range(SHIM_LENGTH)]
            },
        }
    )
    # -fmap:
    #    * Case 1 in https://bids-specification.readthedocs.io/en/stable/04-modality-specific-files/01-magnetic-resonance-imaging-data.html#fieldmap-data
    fmap_struct: dict[str, str | dict[str, list[float]]] = {
        "{p}_acq-case1_{s}.nii.gz".format(p=prefix, s=suffix): ""
        for suffix in ["phasediff", "magnitude1", "magnitude2"]
    }
    expected_fmap_groups = {
        "{p}_acq-case1".format(p=prefix): [
            "{p}_acq-case1_phasediff.json".format(
                p=op.join(session_path, "fmap", prefix)
            )
        ]
    }
    fmap_struct.update(
        {"{p}_acq-case1_phasediff.json".format(p=prefix): {"ShimSetting": dwi_shims}}
    )
    #    * Case 2:
    fmap_struct.update(
        {
            "{p}_acq-case2_{s}.nii.gz".format(p=prefix, s=suffix): ""
            for suffix in ["magnitude1", "magnitude2", "phase1", "phase2"]
        }
    )
    expected_fmap_groups.update(
        {
            "{p}_acq-case2".format(p=prefix): [
                "{p}_acq-case2_phase{n}.json".format(
                    p=op.join(session_path, "fmap", prefix), n=n
                )
                for n in [1, 2]
            ]
        }
    )
    fmap_struct.update(
        {
            "{p}_acq-case2_phase{n}.json".format(p=prefix, n=n): {
                "ShimSetting": func_shims_A
            }
            for n in [1, 2]
        }
    )
    #    * Case 3:
    fmap_struct.update(
        {
            "{p}_acq-case3_{s}.nii.gz".format(p=prefix, s=suffix): ""
            for suffix in ["magnitude", "fieldmap"]
        }
    )
    expected_fmap_groups.update(
        {
            "{p}_acq-case3".format(p=prefix): [
                "{p}_acq-case3_fieldmap.json".format(
                    p=op.join(session_path, "fmap", prefix)
                )
            ]
        }
    )
    fmap_struct.update(
        {"{p}_acq-case3_fieldmap.json".format(p=prefix): {"ShimSetting": func_shims_B}}
    )
    # structure for the full session (init the OrderedDict as a list to preserve order):
    session_struct = OrderedDict(
        [
            ("fmap", fmap_struct),
            ("dwi", dwi_struct),
            ("func", func_struct),
        ]
    )
    # add "_scans.tsv" file to the session_struct
    scans_file_content = generate_scans_tsv(session_struct)
    session_struct2: dict[str, Load] = {
        **session_struct,
        f"{prefix}_scans.tsv": scans_file_content,
    }

    create_tree(session_path, session_struct2)

    # 2) Now, let's create a dict with the fmap groups compatible for each run
    # -dwi: each of the runs (1, 2) is compatible with case1 fmap:
    expected_compatible_fmaps = {
        "{p}_acq-A_run-{r}_dwi.json".format(
            p=op.join(session_path, "dwi", prefix), r=runNo
        ): {
            key: val
            for key, val in expected_fmap_groups.items()
            if key in ["{p}_acq-case1".format(p=prefix)]
        }
        for runNo in [1, 2]
    }
    # -func: acq-A is compatible w/ fmap case2; acq-B w/ fmap case3
    expected_compatible_fmaps.update(
        {
            "{p}_acq-{a}_bold.json".format(
                p=op.join(session_path, "func", prefix), a=acq
            ): {
                key: val
                for key, val in expected_fmap_groups.items()
                if key in ["{p}_acq-case{c}".format(p=prefix, c=caseNo)]
            }
            for caseNo, acq in {"2": "A", "3": "B"}.items()
        }
    )
    # -func (cont): acq-unmatched is empty
    expected_compatible_fmaps.update(
        {
            "{p}_acq-unmatched_bold.json".format(
                p=op.join(session_path, "func", prefix)
            ): {}
        }
    )

    # 3) Now, let's create a dict with what we expect for the "IntendedFor":

    sub_match = re.findall("(sub-([a-zA-Z0-9]*))", session_path)
    sub_str = sub_match[0][0]
    expected_prefix = session_path.split(sub_str)[-1].split(op.sep)[-1]

    # dict, with fmap names as keys and the expected "IntendedFor" as values.
    expected_result: dict[str, Optional[list[str]]] = {
        "{p}_acq-case1_{s}.json".format(p=prefix, s="phasediff"): [
            op.join(
                expected_prefix,
                "dwi",
                "{p}_acq-A_run-{r}_dwi.nii.gz".format(p=prefix, r=r),
            )
            for r in [1, 2]
        ]
    }
    expected_result.update(
        {
            "{p}_acq-case2_phase{n}.json".format(p=prefix, n=n):
            # populate_intended_for writes lists:
            [op.join(expected_prefix, "func", "{p}_acq-A_bold.nii.gz".format(p=prefix))]
            for n in [1, 2]
        }
    )
    expected_result.update(
        {
            "{p}_acq-case3_fieldmap.json".format(p=prefix):
            # populate_intended_for writes lists:
            [op.join(expected_prefix, "func", "{p}_acq-B_bold.nii.gz".format(p=prefix))]
        }
    )

    return (
        session_struct2,
        expected_result,
        expected_fmap_groups,
        expected_compatible_fmaps,
    )


def mkstruct(template: str, *params: list) -> dict[str, str | dict[str, str]]:
    struct: dict[str, str | dict[str, str]] = {
        template.format(*ps, ext="nii.gz"): "" for ps in itertools.product(*params)
    }
    struct.update(
        {template.format(*ps, ext="json"): {} for ps in itertools.product(*params)}
    )
    return struct


# Test cases:
# A) pepolar fmaps with ShimSetting in json files
# B) same, with no ShimSetting
# C) magnitude/phase, with ShimSetting
@pytest.mark.parametrize(
    "simulation_function",
    [
        create_dummy_pepolar_bids_session,
        create_dummy_no_shim_settings_bids_session,
        create_dummy_magnitude_phase_bids_session,
    ],
)
def test_find_fmap_groups(
    tmp_path: Path, simulation_function: Callable[[str], DummySession]
) -> None:
    """Test for find_fmap_groups"""
    folder = op.join(tmp_path, "sub-foo")
    _, _, expected_fmap_groups, _ = simulation_function(folder)
    fmap_groups = find_fmap_groups(op.join(folder, "fmap"))
    assert fmap_groups == expected_fmap_groups


# Test cases:
# A) pepolar fmaps with ShimSetting in json files
# B) same, with no ShimSetting
# C) magnitude/phase, with ShimSetting
@pytest.mark.parametrize(
    "simulation_function, match_param",
    [
        (create_dummy_pepolar_bids_session, "Shims"),
        (create_dummy_no_shim_settings_bids_session, "ModalityAcquisitionLabel"),
        (
            create_dummy_no_shim_settings_custom_label_bids_session,
            "CustomAcquisitionLabel",
        ),
        (create_dummy_magnitude_phase_bids_session, "Shims"),
    ],
)
def test_find_compatible_fmaps_for_run(
    tmp_path: Path, simulation_function: Callable[[str], DummySession], match_param: str
) -> None:
    """
    Test find_compatible_fmaps_for_run.

    Parameters
    ----------
    tmp_path
    simulation_function : function
        function to create the directory tree and expected results
    match_param : str
        matching_parameter for assigning fmaps
    """
    folder = op.join(tmp_path, "sub-foo")
    _, _, expected_fmap_groups, expected_compatible_fmaps = simulation_function(folder)
    for modality in ["anat", "dwi", "func"]:
        for json_file in glob(op.join(folder, modality, "*.json")):
            compatible_fmaps = find_compatible_fmaps_for_run(
                json_file, expected_fmap_groups, matching_parameters=[match_param]
            )
            assert compatible_fmaps == expected_compatible_fmaps[json_file]


# Test two scenarios for each case:
# -study without sessions
# -study with sessions
# Cases:
# A) pepolar fmaps with ShimSetting in json files
# B) same, with no ShimSetting
# C) magnitude/phase, with ShimSetting
@pytest.mark.parametrize(
    "folder, expected_prefix, simulation_function, match_param",
    [
        (folder, expected_prefix, sim_func, mp)
        for folder, expected_prefix in zip(
            ["no_sessions/sub-1", "sessions/sub-1/ses-pre"], ["", "ses-pre"]
        )
        for sim_func, mp in [
            (create_dummy_pepolar_bids_session, "Shims"),
            (create_dummy_no_shim_settings_bids_session, "ModalityAcquisitionLabel"),
            (
                create_dummy_no_shim_settings_custom_label_bids_session,
                "CustomAcquisitionLabel",
            ),
            (create_dummy_magnitude_phase_bids_session, "Shims"),
        ]
    ],
)
def test_find_compatible_fmaps_for_session(
    tmp_path: Path,
    folder: str,
    expected_prefix: str,
    simulation_function: Callable[[str], DummySession],
    match_param: str,
) -> None:
    """
    Test find_compatible_fmaps_for_session.

    Parameters
    ----------
    tmp_path
    folder : str
        path to BIDS study to be simulated, relative to tmp_path
    expected_prefix : str
        expected start of the "IntendedFor" elements
    simulation_function : function
        function to create the directory tree and expected results
    match_param : str
        matching_parameter for assigning fmaps
    """
    session_folder = op.join(tmp_path, folder)
    _, _, _, expected_compatible_fmaps = simulation_function(session_folder)

    compatible_fmaps = find_compatible_fmaps_for_session(
        session_folder, matching_parameters=[match_param]
    )

    assert compatible_fmaps == expected_compatible_fmaps


# Test two scenarios for each case:
# -study without sessions
# -study with sessions
# Cases:
# A) pepolar fmaps with ShimSetting in json files
# B) same, with no ShimSetting
# C) magnitude/phase, with ShimSetting
@pytest.mark.parametrize(
    "folder, expected_prefix, simulation_function",
    [
        (folder, expected_prefix, sim_func)
        for folder, expected_prefix in zip(
            ["no_sessions/sub-1", "sessions/sub-1/ses-pre"], ["", "ses-pre"]
        )
        for sim_func in [
            create_dummy_pepolar_bids_session,
            create_dummy_no_shim_settings_bids_session,
            create_dummy_magnitude_phase_bids_session,
        ]
    ],
)
def test_select_fmap_from_compatible_groups(
    tmp_path: Path,
    folder: str,
    expected_prefix: str,
    simulation_function: Callable[[str], DummySession],
) -> None:
    """Test select_fmap_from_compatible_groups"""
    session_folder = op.join(tmp_path, folder)
    _, _, _, expected_compatible_fmaps = simulation_function(session_folder)

    for json_file, fmap_groups in expected_compatible_fmaps.items():
        for criterion in AllowedCriteriaForFmapAssignment:
            if not op.dirname(json_file).endswith("fmap"):
                selected_fmap = select_fmap_from_compatible_groups(
                    json_file, fmap_groups, criterion=criterion
                )
            # when the criterion is 'First', you should get the first of
            # the compatible_fmaps (for that json_file), if it is 'Closest',
            # it should be the last one (the fmaps are "run" at the
            # beginning of the session)
            if selected_fmap:
                if criterion == "First":
                    assert (
                        selected_fmap == sorted(expected_compatible_fmaps[json_file])[0]
                    )
                elif criterion == "Closest":
                    assert (
                        selected_fmap
                        == sorted(expected_compatible_fmaps[json_file])[-1]
                    )
            else:
                assert not expected_compatible_fmaps[json_file]


# Test two scenarios for each case:
# -study without sessions
# -study with sessions
# Cases:
# A) pepolar fmaps with ShimSetting in json files
# B) same, with no ShimSetting
# C) magnitude/phase, with ShimSetting
@pytest.mark.parametrize(
    "folder, expected_prefix, simulation_function, match_param",
    [
        (folder, expected_prefix, sim_func, mp)
        for folder, expected_prefix in zip(
            ["no_sessions/sub-1", "sessions/sub-1/ses-pre"], ["", "ses-pre"]
        )
        for sim_func, mp in [
            (create_dummy_pepolar_bids_session, "Shims"),
            (create_dummy_no_shim_settings_bids_session, "ModalityAcquisitionLabel"),
            (
                create_dummy_no_shim_settings_custom_label_bids_session,
                "CustomAcquisitionLabel",
            ),
            (create_dummy_magnitude_phase_bids_session, "Shims"),
        ]
    ],
)
def test_populate_intended_for(
    tmp_path: Path,
    folder: str,
    expected_prefix: str,
    simulation_function: Callable[[str], DummySession],
    match_param: str,
) -> None:
    """
    Test populate_intended_for.
    Parameters
    ----------
    tmp_path
    folder : str
        path to BIDS study to be simulated, relative to tmp_path
    expected_prefix : str
        expected start of the "IntendedFor" elements
    simulation_function : function
        function to create the directory tree and expected results
    match_param : str
        matching_parameter for assigning fmaps
    """

    session_folder = op.join(tmp_path, folder)
    session_struct, expected_result, _, _ = simulation_function(session_folder)
    populate_intended_for(
        session_folder, matching_parameters=match_param, criterion="First"
    )

    # Now, loop through the jsons in the fmap folder and make sure it matches
    # the expected result:
    fmap_folder = op.join(session_folder, "fmap")
    fmap = session_struct["fmap"]
    assert isinstance(fmap, dict)
    for j in fmap.keys():
        if j.endswith(".json"):
            assert j in expected_result.keys()
            data = load_json(op.join(fmap_folder, j))
            if expected_result[j]:
                assert data["IntendedFor"] == expected_result[j]
                # Also, make sure the run with random shims is not here:
                # (It is assured by the assert above, but let's make it
                # explicit)
                run_prefix = j.split("_acq")[0]
                assert (
                    "{p}_acq-unmatched_bold.nii.gz".format(p=run_prefix)
                    not in data["IntendedFor"]
                )
            else:
                assert "IntendedFor" not in data.keys()


def test_BIDSFile() -> None:
    """Tests for the BIDSFile class"""

    # define entities in the correct order:
    sorted_entities = [
        ("sub", "Jason"),
        ("acq", "Treadstone"),
        ("run", "2"),
        ("echo", "1"),
    ]
    # 'sub-Jason_acq-Treadstone_run-2_echo-1':
    expected_sorted_str = "_".join(["-".join(e) for e in sorted_entities])
    # expected BIDSFile:
    suffix = "T1w"
    extension = "nii.gz"
    expected_bids_file = BIDSFile(OrderedDict(sorted_entities), suffix, extension)

    # entities in random order:
    idcs = list(range(len(sorted_entities)))
    shuffle(idcs)
    shuffled_entities = [sorted_entities[i] for i in idcs]
    shuffled_str = "_".join(["-".join(e) for e in shuffled_entities])

    # Test __eq__ method.
    # It should consider equal BIDSFiles with the same entities even in different order:
    assert (
        BIDSFile(OrderedDict(shuffled_entities), suffix, extension)
        == expected_bids_file
    )

    # Test __getitem__:
    assert all([expected_bids_file[k] == v for k, v in shuffled_entities])

    # Test filename parser and  __str__ method:
    # Note: the __str__ method should return entities in the correct order
    ending = "_T1w.nii.gz"  # suffix + extension
    my_bids_file = BIDSFile.parse(shuffled_str + ending)
    assert my_bids_file == expected_bids_file
    assert str(my_bids_file) == expected_sorted_str + ending

    ending = ".json"  # just extension
    my_bids_file = BIDSFile.parse(shuffled_str + ending)
    assert my_bids_file.suffix == ""
    assert str(my_bids_file) == expected_sorted_str + ending

    ending = "_T1w"  # just suffix
    my_bids_file = BIDSFile.parse(shuffled_str + ending)
    assert my_bids_file.extension is None
    assert str(my_bids_file) == expected_sorted_str + ending

    # Complain if entity 'sub' is not set:
    with pytest.raises(ValueError) as e_info:
        assert str(BIDSFile.parse("dir-reversed.json"))
        assert "sub-" in str(e_info.value)

    # Test set method:
    # -for a new entity, just set it without a complaint:
    my_bids_file["dir"] = "AP"
    assert my_bids_file["dir"] == "AP"
    # -for an existing entity, don't change it by default:
    my_bids_file["echo"] = "2"
    assert (
        my_bids_file["echo"] == expected_bids_file["echo"]
    )  # still the original value
    # -for an existing entity, you can overwrite it with "set":
    my_bids_file.set("echo", "2")
    assert my_bids_file["echo"] == "2"


@pytest.mark.skipif(not have_datalad, reason="no datalad")
def test_ME_mag_phase_conversion(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    subID: str = "MEGRE",
    heuristic: str = "bids_ME.py",
) -> None:
    """Unit test for the case of multi-echo GRE data with
    magnitude and phase.
    The different echoes should be labeled automatically.
    """
    monkeypatch.chdir(tmp_path)
    try:
        datadir = fetch_data(tmp_path, f"dicoms/velasco/{subID}")
    except IncompleteResultsError as exc:
        pytest.skip("Failed to fetch test data: %s" % str(exc))

    outdir = tmp_path / "out"
    outdir.mkdir()
    args = gen_heudiconv_args(datadir, str(outdir), subID, heuristic)
    runner(args)  # run conversion

    # Check that the expected files have been extracted.
    # This also checks that the "echo" entity comes before "part":
    for part in ["mag", "phase"]:
        for e in range(1, 4):
            for ext in ["nii.gz", "json"]:
                assert op.exists(
                    op.join(outdir, "sub-%s", "anat", "sub-%s_echo-%s_part-%s_MEGRE.%s")
                    % (subID, subID, e, part, ext)
                )


def test_sanitize_label() -> None:
    assert sanitize_label("az XZ-@09") == "azXZ09"
    with pytest.raises(ValueError):
        sanitize_label(" @ ")



from __future__ import annotations

from datetime import datetime
import json
from json.decoder import JSONDecodeError
import os
import os.path as op
from pathlib import Path
from typing import IO, Any
from unittest.mock import patch

import pytest

from heudiconv.utils import (
    create_tree,
    get_datetime,
    get_heuristic_description,
    get_known_heuristics_with_descriptions,
    json_dumps_pretty,
    load_heuristic,
    load_json,
    remove_prefix,
    remove_suffix,
    save_json,
    strptime_micr,
    update_json,
)

from .utils import HEURISTICS_PATH


def test_get_known_heuristics_with_descriptions() -> None:
    d = get_known_heuristics_with_descriptions()
    assert {"reproin", "convertall"}.issubset(d)
    # ATM we include all, not only those two
    assert len(d) > 2
    assert len(d["reproin"]) > 50  # it has a good one
    assert len(d["reproin"].split(os.sep)) == 1  # but just one line


def test_get_heuristic_description() -> None:
    desc = get_heuristic_description("reproin", full=True)
    assert len(desc) > 1000
    # and we describe such details as
    assert "_ses-" in desc
    assert "_run-" in desc
    # and mention ReproNim ;)
    assert "ReproNim" in desc


def test_load_heuristic() -> None:
    by_name = load_heuristic("reproin")
    from_file = load_heuristic(op.join(HEURISTICS_PATH, "reproin.py"))

    assert by_name
    assert by_name.filename == from_file.filename

    with pytest.raises(ImportError):
        load_heuristic("unknownsomething")

    with pytest.raises(ImportError):
        load_heuristic(op.join(HEURISTICS_PATH, "unknownsomething.py"))


def test_json_dumps_pretty() -> None:
    pretty = json_dumps_pretty
    assert (
        pretty({"SeriesDescription": "Trace:Nov 13 2017 14-36-14 EST"})
        == '{\n  "SeriesDescription": "Trace:Nov 13 2017 14-36-14 EST"\n}'
    )
    assert pretty({}) == "{}"
    assert (
        pretty({"a": -1, "b": "123", "c": [1, 2, 3], "d": ["1.0", "2.0"]})
        == '{\n  "a": -1,\n  "b": "123",\n  "c": [1, 2, 3],\n  "d": ["1.0", "2.0"]\n}'
    )
    assert (
        pretty({"a": ["0.3", "-1.9128906358217845e-12", "0.2"]})
        == '{\n  "a": ["0.3", "-1.9128906358217845e-12", "0.2"]\n}'
    )
    # original, longer string
    tstr = (
        "f9a7d4be-a7d7-47d2-9de0-b21e9cd10755||"
        "Sequence: ve11b/master r/50434d5; "
        "Mar  3 2017 10:46:13 by eja"
    )
    # just the date which reveals the issue
    # tstr = 'Mar  3 2017 10:46:13 by eja'
    assert pretty({"WipMemBlock": tstr}) == '{\n  "WipMemBlock": "%s"\n}' % tstr


def test_load_json(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    # test invalid json
    ifname = "invalid.json"
    invalid_json_file = str(tmp_path / ifname)
    create_tree(str(tmp_path), {ifname: "I'm Jason Bourne"})

    with pytest.raises(JSONDecodeError):
        load_json(str(invalid_json_file))

    # and even if we ask to retry a few times -- should be the same
    with pytest.raises(JSONDecodeError):
        load_json(str(invalid_json_file), retry=3)

    with pytest.raises(FileNotFoundError):
        load_json("absent123not.there", retry=3)

    assert ifname in caplog.text

    # test valid json
    vcontent = {"secret": "spy"}
    vfname = "valid.json"
    valid_json_file = str(tmp_path / vfname)
    save_json(valid_json_file, vcontent)

    assert load_json(valid_json_file) == vcontent

    calls = [0]
    json_load = json.load

    def json_load_patched(fp: IO[str]) -> Any:
        calls[0] += 1
        if calls[0] == 1:
            # just reuse bad file
            load_json(str(invalid_json_file))
        elif calls[0] == 2:
            raise FileNotFoundError()
        else:
            return json_load(fp)

    with patch.object(json, "load", json_load_patched):
        assert load_json(valid_json_file, retry=3) == vcontent


def test_update_json(tmp_path: Path) -> None:
    """
    Test utils.update_json()
    """
    dummy_json_file = str(tmp_path / "dummy.json")
    some_content = {"name": "Jason", "age": 30, "city": "New York"}
    save_json(dummy_json_file, some_content, pretty=True)

    added_content = {
        "LastName": "Bourne",
        "Movies": [
            "The Bourne Identity",
            "The Bourne Supremacy",
            "The Bourne Ultimatum",
            "The Bourne Legacy",
            "Jason Bourne",
        ],
    }
    update_json(dummy_json_file, added_content)

    # check that it was added:
    with open(dummy_json_file) as f:
        data = json.load(f)
    some_content.update(added_content)
    assert data == some_content


def test_get_datetime() -> None:
    """
    Test utils.get_datetime()
    """
    assert get_datetime("20200512", "162130") == "2020-05-12T16:21:30"
    assert get_datetime("20200512", "162130.5") == "2020-05-12T16:21:30.500000"
    assert (
        get_datetime("20200512", "162130.5", microseconds=False)
        == "2020-05-12T16:21:30"
    )


@pytest.mark.parametrize(
    "dt, fmt",
    [
        ("20230310190100", "%Y%m%d%H%M%S"),
        ("2023-04-02T11:47:09", "%Y-%m-%dT%H:%M:%S"),
    ],
)
def test_strptime_micr(dt: str, fmt: str) -> None:
    target = datetime.strptime(dt, fmt)
    assert strptime_micr(dt, fmt) == target
    assert strptime_micr(dt, fmt + "[.%f]") == target
    assert strptime_micr(dt + ".0", fmt + "[.%f]") == target
    assert strptime_micr(dt + ".000000", fmt + "[.%f]") == target
    assert strptime_micr(dt + ".1", fmt + "[.%f]") == datetime.strptime(
        dt + ".1", fmt + ".%f"
    )


def test_remove_suffix() -> None:
    """
    Test utils.remove_suffix()
    """
    s = "jason.bourne"
    assert remove_suffix(s, "") == s
    assert remove_suffix(s, "foo") == s
    assert remove_suffix(s, ".bourne") == "jason"


def test_remove_prefix() -> None:
    """
    Test utils.remove_prefix()
    """
    s = "jason.bourne"
    assert remove_prefix(s, "") == s
    assert remove_prefix(s, "foo") == s
    assert remove_prefix(s, "jason") == ".bourne"



__author__ = "HeuDiConv team and contributors"
__url__ = "https://github.com/nipy/heudiconv"
__packagename__ = "heudiconv"
__description__ = "Heuristic DICOM Converter"
__license__ = "Apache 2.0"
__longdesc__ = """Convert DICOM dirs based on heuristic info - HeuDiConv
uses the dcmstack package and dcm2niix tool to convert DICOM directories or
tarballs into collections of NIfTI files following pre-defined heuristic(s)."""

CLASSIFIERS = [
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Topic :: Scientific/Engineering",
    "Typing :: Typed",
]

PYTHON_REQUIRES = ">=3.8"

REQUIRES = [
    # not usable in some use cases since might be just a downloader, not binary
    # 'dcm2niix',
    "dcmstack>=0.8",
    "etelemetry",
    "filelock>=3.0.12",
    "nibabel",
    "nipype >=1.2.3",
    "pydicom >= 1.0.0",
]

TESTS_REQUIRES = [
    "pytest",
    "tinydb",
    "inotify",
]

MIN_DATALAD_VERSION = "0.13.0"
EXTRA_REQUIRES = {
    "tests": TESTS_REQUIRES,
    "extras": [
        "duecredit",  # optional dependency
    ],  # Requires patched version ATM ['dcmstack'],
    "datalad": ["datalad >=%s" % MIN_DATALAD_VERSION],
}

# Flatten the lists
EXTRA_REQUIRES["all"] = sum(EXTRA_REQUIRES.values(), [])



#
# Tests for reproin.py
#
from __future__ import annotations

import re
from typing import NamedTuple
from unittest.mock import patch

import pytest

from . import reproin
from .reproin import (
    filter_files,
    fix_canceled_runs,
    fix_dbic_protocol,
    fixup_subjectid,
    get_dups_marked,
    get_unique,
    md5sum,
    parse_series_spec,
    sanitize_str,
)


class FakeSeqInfo(NamedTuple):
    accession_number: str
    study_description: str
    field1: str
    field2: str


def test_get_dups_marked() -> None:
    no_dups: dict[tuple[str, tuple[str, ...], None], list[int]] = {
        ("some", ("foo",), None): [1]
    }
    assert get_dups_marked(no_dups) == no_dups

    info: dict[tuple[str, tuple[str, ...], None], list[int | str]] = {
        ("bu", ("du",), None): [1, 2],
        ("smth", (), None): [3],
        ("smth2", ("apple", "banana"), None): ["a", "b", "c"],
    }

    assert (
        get_dups_marked(info)
        == get_dups_marked(info, True)
        == {
            ("bu__dup-01", ("du",), None): [1],
            ("bu", ("du",), None): [2],
            ("smth", (), None): [3],
            ("smth2__dup-01", ("apple", "banana"), None): ["a"],
            ("smth2__dup-02", ("apple", "banana"), None): ["b"],
            ("smth2", ("apple", "banana"), None): ["c"],
        }
    )

    assert get_dups_marked(info, per_series=False) == {
        ("bu__dup-01", ("du",), None): [1],
        ("bu", ("du",), None): [2],
        ("smth", (), None): [3],
        ("smth2__dup-02", ("apple", "banana"), None): ["a"],
        ("smth2__dup-03", ("apple", "banana"), None): ["b"],
        ("smth2", ("apple", "banana"), None): ["c"],
    }


def test_filter_files() -> None:
    # Filtering is currently disabled -- any sequence directory is Ok
    assert filter_files("/home/mvdoc/dbic/09-run_func_meh/0123432432.dcm")
    assert filter_files("/home/mvdoc/dbic/run_func_meh/012343143.dcm")


def test_md5sum() -> None:
    assert md5sum("cryptonomicon") == "1cd52edfa41af887e14ae71d1db96ad1"
    assert md5sum("mysecretmessage") == "07989808231a0c6f522f9d8e34695794"


def test_fix_canceled_runs() -> None:
    class FakeSeqInfo(NamedTuple):
        accession_number: str
        series_id: str
        protocol_name: str
        series_description: str

    seqinfo: list[FakeSeqInfo] = []
    runname = "func_run+"
    for i in range(1, 6):
        seqinfo.append(
            FakeSeqInfo("accession1", "{0:02d}-".format(i) + runname, runname, runname)
        )

    fake_accession2run = {"accession1": ["^01-", "^03-"]}

    with patch.object(reproin, "fix_accession2run", fake_accession2run):
        seqinfo_ = fix_canceled_runs(seqinfo)  # type: ignore[arg-type]

    for i, s in enumerate(seqinfo_, 1):
        output = runname
        if i == 1 or i == 3:
            output = "cancelme_" + output
        for key in ["series_description", "protocol_name"]:
            value = getattr(s, key)
            assert value == output
        # check we didn't touch series_id
        assert s.series_id == "{0:02d}-".format(i) + runname


def test_fix_dbic_protocol() -> None:
    accession_number = "A003"
    seq1 = FakeSeqInfo(
        accession_number,
        "mystudy",
        "02-anat-scout_run+_MPR_sag",
        "11-func_run-life2_acq-2mm692",
    )
    seq2 = FakeSeqInfo(accession_number, "mystudy", "nochangeplease", "nochangeeither")

    seqinfos = [seq1, seq2]
    protocols2fix = {
        md5sum("mystudy"): [
            (r"scout_run\+", "THESCOUT-runX"),
            ("run-life[0-9]", "run+_task-life"),
        ],
        re.compile("^my.*"): [("THESCOUT-runX", "THESCOUT")],
        # rely on 'catch-all' to fix up above scout
        "": [("THESCOUT", "scout")],
    }

    with patch.object(reproin, "protocols2fix", protocols2fix), patch.object(
        reproin, "series_spec_fields", ["field1"]
    ):
        seqinfos_ = fix_dbic_protocol(seqinfos)  # type: ignore[arg-type]
    assert seqinfos[1] == seqinfos_[1]  # type: ignore[comparison-overlap]
    # field2 shouldn't have changed since I didn't pass it
    assert seqinfos_[0] == FakeSeqInfo(  # type: ignore[comparison-overlap]
        accession_number, "mystudy", "02-anat-scout_MPR_sag", seq1.field2
    )

    # change also field2 please
    with patch.object(reproin, "protocols2fix", protocols2fix), patch.object(
        reproin, "series_spec_fields", ["field1", "field2"]
    ):
        seqinfos_ = fix_dbic_protocol(seqinfos)  # type: ignore[arg-type]
    assert seqinfos[1] == seqinfos_[1]  # type: ignore[comparison-overlap]
    # now everything should have changed
    assert seqinfos_[0] == FakeSeqInfo(  # type: ignore[comparison-overlap]
        accession_number,
        "mystudy",
        "02-anat-scout_MPR_sag",
        "11-func_run+_task-life_acq-2mm692",
    )


def test_sanitize_str() -> None:
    assert sanitize_str("super@duper.faster") == "superduperfaster"
    assert sanitize_str("perfect") == "perfect"
    assert sanitize_str("never:use:colon:!") == "neverusecolon"


def test_fixupsubjectid() -> None:
    assert fixup_subjectid("abra") == "abra"
    assert fixup_subjectid("sub") == "sub"
    assert fixup_subjectid("sid") == "sid"
    assert fixup_subjectid("sid000030") == "sid000030"
    assert fixup_subjectid("sid0000030") == "sid000030"
    assert fixup_subjectid("sid00030") == "sid000030"
    assert fixup_subjectid("sid30") == "sid000030"
    assert fixup_subjectid("SID30") == "sid000030"


def test_parse_series_spec() -> None:
    pdpn = parse_series_spec

    assert pdpn("nondbic_func-bold") == {}
    assert pdpn("cancelme_func-bold") == {}

    assert (
        pdpn("bids_func-bold")
        == pdpn("func-bold")
        == {"datatype": "func", "datatype_suffix": "bold"}
    )

    # pdpn("bids_func_ses+_task-boo_run+") == \
    # order and PREFIX: should not matter, as well as trailing spaces
    assert (
        pdpn(" PREFIX:bids_func_ses+_task-boo_run+  ")
        == pdpn("PREFIX:bids_func_ses+_task-boo_run+")
        == pdpn("WIP func_ses+_task-boo_run+")
        == pdpn("bids_func_ses+_run+_task-boo")
        == {
            "datatype": "func",
            # 'datatype_suffix': 'bold',
            "session": "+",
            "run": "+",
            "task": "boo",
        }
    )

    # TODO: fix for that
    assert (
        pdpn("bids_func-pace_ses-1_task-boo_acq-bu_bids-please_run-2__therest")
        == pdpn("bids_func-pace_ses-1_run-2_task-boo_acq-bu_bids-please__therest")
        == pdpn("func-pace_ses-1_task-boo_acq-bu_bids-please_run-2")
        == {
            "datatype": "func",
            "datatype_suffix": "pace",
            "session": "1",
            "run": "2",
            "task": "boo",
            "acq": "bu",
            "bids": "bids-please",
        }
    )

    assert pdpn("bids_anat-scout_ses+") == {
        "datatype": "anat",
        "datatype_suffix": "scout",
        "session": "+",
    }

    assert pdpn("anat_T1w_acq-MPRAGE_run+") == {
        "datatype": "anat",
        "run": "+",
        "acq": "MPRAGE",
        "datatype_suffix": "T1w",
    }

    # Check for currently used {date}, which should also should get adjusted
    # from (date) since Philips does not allow for {}
    assert (
        pdpn("func_ses-{date}")
        == pdpn("func_ses-(date)")
        == {"datatype": "func", "session": "{date}"}
    )

    assert pdpn("fmap_dir-AP_ses-01") == {
        "datatype": "fmap",
        "session": "01",
        "dir": "AP",
    }


def test_get_unique() -> None:
    accession_number = "A003"
    acqs = [
        FakeSeqInfo(accession_number, "mystudy", "nochangeplease", "nochangeeither"),
        FakeSeqInfo(accession_number, "mystudy2", "nochangeplease", "nochangeeither"),
    ]

    assert get_unique(acqs, "accession_number") == accession_number  # type: ignore[arg-type]
    with pytest.raises(AssertionError) as ce:
        get_unique(acqs, "study_description")  # type: ignore[arg-type]
    assert (
        str(ce.value)
        == "Was expecting a single value for attribute 'study_description' but got: mystudy, mystudy2"
    )



import os

import pytest


@pytest.fixture(autouse=True, scope="session")
def git_env() -> None:
    os.environ["GIT_AUTHOR_EMAIL"] = "maxm@example.com"
    os.environ["GIT_AUTHOR_NAME"] = "Max Mustermann"
    os.environ["GIT_COMMITTER_EMAIL"] = "maxm@example.com"
    os.environ["GIT_COMMITTER_NAME"] = "Max Mustermann"
